VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "AGILogic"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Attribute VB_Ext_KEY = "SavedWithClassBuilder6" ,"Yes"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E067277007C"
Attribute VB_Ext_KEY = "Member0" ,"SourceCode"
Option Explicit
Option Compare Text
Implements AGIResource


Private WithEvents agRes As AGIResource
Attribute agRes.VB_VarHelpID = -1
Private mResID As String
Private mDescription As String

Private mIsDirty As Boolean
Private mWriteProps As Boolean

Private WithEvents mSourceCode As SourceCode
Attribute mSourceCode.VB_VarHelpID = -1
Private mCompiledCRC As Long
Private mIsRoom As Boolean

'other
Private strErrSource As String

Public Property Let IsRoom(ByVal NewVal As Boolean)
  
  'if in a game, and logic 0
  If agRes.InGame And agRes.Number = 0 Then
    On Error GoTo 0: Err.Raise 450, "Can't assign to read-only property"
    Exit Property
  End If
  
  'if changing
  If mIsRoom <> NewVal Then
    mIsRoom = NewVal
    
    WriteProperty PC_LOGIC, PT_ROOM, agRes.Number, CStr(mIsRoom)
  End If
End Property

Public Property Get IsRoom() As Boolean
  
  'if logic 0 in a game
  If agRes.InGame And agRes.Number = 0 Then
    IsRoom = False
  Else
    IsRoom = mIsRoom
  End If
End Property

Private Sub SaveProps()

  'save properties
  WriteProperty PC_LOGIC, PT_ID, agRes.Number, mResID
  WriteProperty PC_LOGIC, PT_DESC, agRes.Number, mDescription
  'only write syntax if known
  If mSourceCode.Syntax <> lsUnknown Then
    WriteProperty PC_LOGIC, PT_SYNTAX, agRes.Number, CStr(mSourceCode.Syntax)
  End If
  WriteProperty PC_LOGIC, PT_CRC32, agRes.Number, Hex$(mCompiledCRC)
  WriteProperty PC_LOGIC, PT_ROOM, agRes.Number, CStr(mIsRoom)
  
  'set flag indicating all props are uptodate
  mWriteProps = False
End Sub

Friend Sub SetCRC()

  'called after compilation to force CRC for this logic
  
  
  'get new source crc
  mCompiledCRC = CRC32(StrConv(mSourceCode.Text, vbFromUnicode))
  
  'set writeprop flag so new crc value is saved
  mWriteProps = True
  'save props
  SaveProps
End Sub

Friend Sub SetFlags(ByRef IsDirty As Boolean, ByRef WriteProps As Boolean)
  
  IsDirty = mIsDirty
  WriteProps = mWriteProps
End Sub


Public Property Get Compiled() As Boolean

  'return true if source code CRC is same as
  'stored compiled crc
  
  'if source is loaded,
  If mSourceCode.Loaded Then
    Compiled = (mSourceCode.CRC = mCompiledCRC)
  Else
    'if source not loaded, always return false
    Compiled = False
  End If
End Property

Friend Property Let CompiledCRC(NewCRC As Long)
  'only called by GetGameProperty method
  mCompiledCRC = NewCRC
  
End Property

Public Property Get CompiledCRC() As Long
  
  On Error Resume Next
  
  'if not loaded
  If Not agRes.Loaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'if not in a game
  If Not agRes.InGame Then
    'no crc
    CompiledCRC = 0
  Else
    'if no crc
    If mCompiledCRC = 0 Then
      'raise error
      On Error GoTo 0: Err.Raise vbObjectError + 592, strErrSource, LoadResString(592)
      Exit Property
    End If
  End If
  
  CompiledCRC = mCompiledCRC
End Property
Public Sub SetLogic(NewLogic As AGILogic)
  'creates a copy of NewLogic into this logic
  
  Dim i As Long
  
  On Error GoTo ErrHandler
  
  With NewLogic
    'add WinAGI items
    mResID = .ID
    mDescription = .Description
    mIsRoom = .IsRoom
    
    'if loaded,
    If .Loaded Then
      mSourceCode.SetParent Me
    End If
    mSourceCode.SetCode .SourceCode
    mSourceCode.IsDirty = .SourceCode.IsDirty
    
    agRes.SetRes .Resource
    
    'if resource loaded,
    If agRes.Loaded Then
      'add resource data
      agRes.AllData = .Resource.AllData
    End If
        
    'copy CRC AFTER setting resource data
    'because the resource change event resets CRC value
    mCompiledCRC = .CompiledCRC
    
    'copy dirty flag and writeprop flag
    .SetFlags mIsDirty, mWriteProps
  End With
Exit Sub

ErrHandler:
  'if error is NOT the no crc error,
  If Err.Number <> vbObjectError + 592 Then
    Debug.Assert False
  End If
  Resume Next
End Sub



Public Sub Clear()
  'clears out source code text
  'and clears the resource data
  
  On Error GoTo ErrHandler
  
  If Not agRes.Loaded Then
    'nothing to clear
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'clear resource
  agRes.Clear
  'write end of logic code (&H0)
  agRes.WriteByte 0

  'if source is loaded,
  If mSourceCode.Loaded Then
    'clear the source code by setting it to
    ''return' command
    Select Case mSourceCode.Syntax
    Case lsVB
      mSourceCode.Text = agCmds(0).Name & "()" & vbNewLine & vbNewLine & "'Messages"
    Case lsAGI
      mSourceCode.Text = agCmds(0).Name & "();" & vbNewLine & vbNewLine & "//messages"
    End Select
  
    'reset crc
    mCompiledCRC = CRC32(StrConv(mSourceCode.Text, vbFromUnicode))
  End If
  
  'note change by marking resource as dirty
  mIsDirty = True
  mSourceCode.IsDirty = True
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub
Public Property Get Description() As String
  Description = mDescription
End Property

Public Property Let Description(ByVal NewDescription As String)

  'limit description to 1K
  NewDescription = Left$(NewDescription, 1024)
  
  'if changing
  If NewDescription <> mDescription Then
    mDescription = NewDescription
    
    If agRes.InGame Then
      WriteProperty PC_LOGIC, PT_DESC, agRes.Number, mDescription
    End If
  End If
End Property


Public Sub Export(ExportFile As String, Optional ByVal ResetDirty As Boolean = True)

  Dim intFile As Integer
  Dim strTempFile As String
  
  'if not loaded
  If Not agRes.Loaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'if not in a game,
  If Not agRes.InGame Then
    'not allowed; nothing to export
    On Error GoTo 0: Err.Raise vbObjectError + 668, strErrSource, LoadResString(668)
  End If
  
  On Error GoTo ErrHandler
  
  'export a logic resource file
  agRes.Export ExportFile
Exit Sub

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Sub


Public Sub Import(ImportFile As String, Optional ByVal AsSource As Boolean = False)
  'imports a logic resource
  'import can be a compiled logic, or source code
  
  On Error GoTo ErrHandler
  
  'set ID to match file name
  '(if logic is in a game, the id will be reset later)
  mResID = JustFileName(ImportFile)
  If Len(mResID) > 64 Then
    mResID = Left$(mResID, 64)
  End If
  
  'create new resource (which marks the resource as loaded)
  agRes.NewResource
  
  'if importing source code
  If AsSource Then
    'set parent for sourcecode
    mSourceCode.SetParent Me
  
    'load the source code
    mSourceCode.Load ImportFile
  Else
    'import the compiled resource
    agRes.Import ImportFile
    
    'if there was an error,
    If Err.Number <> 0 Then
      lngError = Err.Number
      strErrSrc = Err.Source
      strError = Err.Description
      
      Unload
      On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
      Exit Sub
    End If
    
    'set parent for sourcecode
    mSourceCode.SetParent Me
    
    On Error Resume Next
    'load the source code by decompiling
    mSourceCode.Load vbNullString, True
    
    If Err.Number <> 0 Then
      'pass along error (by exiting without clearing it)
      Exit Sub
    End If
  End If
  
  'reset dirty flags
  mIsDirty = False
  mSourceCode.IsDirty = False
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub

Friend Sub Init(ByVal ResNum As Byte)
  
  'initialize resource file
  agRes.Init ResNum
  
  'if res is zero
  If ResNum = 0 Then
    'reset isroom flag
    mIsRoom = False
  End If
End Sub
Public Property Get Loaded() As Boolean

  Loaded = agRes.Loaded
End Property

Public Property Get ID() As String
  
  ID = mResID
End Property


Public Property Let ID(ByVal NewID As String)
  'sets the ID for a resource;
  'resource IDs must be unique to each resource type
  'max length of ID is 64 characters
  'min of 1 character
  
  Dim tmpLog As AGILogic
  
  Err.Clear
  On Error GoTo ErrHandler
  
  'validate length
  If LenB(NewID) = 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 667, strErrSource, LoadResString(667)
    Exit Property
  ElseIf Len(NewID) > 64 Then
    NewID = Left$(NewID, 64)
  End If
  
  'if changing,
  If NewID <> mResID Then
    'if in a game,
    If agRes.InGame Then
      'step through other resources
      For Each tmpLog In agLogs
        'if resource IDs are same
        If tmpLog.ID = NewID Then
          'if not the same logic
          If tmpLog.Number <> agRes.Number Then
            'error
            On Error GoTo 0: Err.Raise vbObjectError + 623, strErrSource, LoadResString(623)
            Exit Property
          End If
        End If
      Next
      'source file always tracks ID for ingame resources
      mSourceCode.ChangeSourceFile NewID
    End If
  
    'save ID
    mResID = NewID
    If agRes.InGame Then
      WriteProperty PC_LOGIC, PT_ID, agRes.Number, mResID
    End If
    
    'reset compiler list of ids
    blnSetIDs = False
  End If
Exit Property

ErrHandler:
  lngError = Err.Number
  strErrSrc = Err.Source
  strError = Err.Description
  
  Debug.Assert False
  On Error GoTo 0: Err.Raise vbObjectError + 686, strErrSrc, Replace(Replace(LoadResString(686), ARG1, "Logic"), ARG2, strError)
End Property
Friend Property Let FriendID(NewID As String)

  'sets the ID for a resource internally
  'does not validate the ID or set dirty flag
  
  'save ID
  mResID = NewID
End Property
Friend Property Let IsDirty(ByVal NewDirtyState As Boolean)
  
  mIsDirty = NewDirtyState
End Property

Public Sub Load()
  
  On Error Resume Next
  
  Debug.Assert agRes.InGame
  
  'if resource data not loaded yet,
  If Not agRes.Loaded Then
    'load resource
    agRes.Load
    If Err.Number <> 0 Then
      'pass along the error (by exiting without clearing it)
      'pass error along
      lngError = Err.Number
      strError = Err.Description
      strErrSrc = Err.Source
      
      On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
      Exit Sub
    End If
  
    'set parent for sourcecode
    mSourceCode.SetParent Me
    
    'if there is a CRC value, get it
    mCompiledCRC = CLng(Val("&H" & ReadProperty(PC_LOGIC, PT_CRC32, agRes.Number)))
    If mCompiledCRC = 0 Then
      'no crc; use a single space string
      mCompiledCRC = CRC32(StrConv(" ", vbFromUnicode))
    End If
    
    'clear dirty flag
    mIsDirty = False
  End If
End Sub

Public Sub Compile()
  'compiles the source code for this logic
  'and saves the resource
  On Error Resume Next
  
  'if not in a game
  If Not agRes.InGame Then
    On Error GoTo 0: Err.Raise vbObjectError + 618, strErrSource, LoadResString(618)
    Exit Sub
  End If
  
  'if source code not loaded,
  If Not mSourceCode.Loaded Then
    On Error GoTo 0: Err.Raise vbObjectError + 603, strErrSource, LoadResString(603)
    Exit Sub
  End If
    
  'if no data in sourcecode
  If LenB(mSourceCode.Text) = 0 Then
    On Error GoTo 0: Err.Raise vbObjectError + 546, strErrSource, LoadResString(546)
    Exit Sub
  End If

  'if syntax is not known,
  If mSourceCode.Syntax = lsUnknown Then
    mSourceCode.Syntax = GuessSyntax(mSourceCode.Text, agMainLogSettings.DefaultSyntax)
  End If
  
  'compile the logic
  CompileLogic Me
  
  'save error info
  With Err
    lngError = .Number
    strError = .Description
    strErrSrc = .Source
  End With
  
  'if no error or warning returned
  If lngError = 0 Or lngError = vbObjectError + 636 Then
    'set dirty flag
    mIsDirty = True
    
    'save logic
    Save
  End If
  
  'pass along errors and/or warnings
  If lngError <> 0 Then
    'pass it along
    On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
  End If
End Sub

Public Property Get Number() As Byte
  Number = agRes.Number
End Property


Public Property Get Resource() As AGIResource
  Set Resource = agRes
End Property

Public Property Get SourceCode() As SourceCode
  'returns the sourcecode associated with this logic
  
'if notloaded,
If Not agRes.Loaded Then
  On Error GoTo 0: Err.Raise vbObjectError + 678, "logics", LoadResString(678)
  Exit Property
End If

  Set SourceCode = mSourceCode
End Property



Public Sub Save(Optional SaveFile As String)
  'this file saves the logic resource to next available VOL file
  
  'NOTE: this saves the compiled resource NOT the
  'text based source code
  'SaveSource saves the source code
  'if Save method is called for a resource NOT in a game,
  'it calls the SaveSource method automatically
  Dim i As Long, tmpVal As String
  
  On Error Resume Next
  
  'if properties need to be written
  If mWriteProps And agRes.InGame Then
    SaveProps
  End If
  
  'if not loaded
  If Not agRes.Loaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'if in a game
  If agRes.InGame Then
    'if dirty
    If mIsDirty Then
      'use the resource save method
      agRes.Save SaveFile
      'if any errors
      If Err.Number <> 0 Then
        'pass error along
        lngError = Err.Number
        strError = Err.Description
        strErrSrc = Err.Source
        
        On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
        Exit Sub
      End If
    End If
  Else
    'if source is dirty
    If mSourceCode.IsDirty Then
      'same as savesource
      mSourceCode.Save SaveFile
    End If
  End If
  
  'reset dirty flag
  mIsDirty = False
End Sub

Public Property Get IsDirty() As Boolean

  'if in a game,
  If agRes.InGame Then
    'if resource is dirty, or prop values need writing,
    IsDirty = (mIsDirty Or mWriteProps)
  Else
    IsDirty = mSourceCode.IsDirty
  End If
  
End Property





Public Sub Unload()
  'unload resource
  agRes.Unload
  mIsDirty = False
  mSourceCode.Unload
  mCompiledCRC = 0
  mSourceCode.SetParent Nothing
End Sub


Friend Property Let WritePropState(ByVal NewWritePropState As Boolean)

  mWriteProps = NewWritePropState
End Property

Private Property Get AGIResource_AllData() As Byte()
  AGIResource_AllData = agRes.AllData
End Property


Private Sub AGIResource_Clear()
  Clear
End Sub


Private Property Let AGIResource_Data(ByVal Pos As Long, ByVal NewData As Byte)
  agRes.Data(Pos) = NewData

End Property

Private Property Get AGIResource_Data(ByVal Pos As Long) As Byte
  On Error GoTo ErrHandler
  
  AGIResource_Data = agRes.Data(Pos)
Exit Property

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Property


Private Property Get AGIResource_EORes() As Boolean
  AGIResource_EORes = agRes.EORes
End Property


Private Sub AGIResource_Export(ExportFile As String)
  agRes.Export ExportFile
End Sub



Private Function AGIResource_GetPos() As Long
  AGIResource_GetPos = agRes.GetPos
End Function


Private Property Get AGIResource_InGame() As Boolean
  AGIResource_InGame = agRes.InGame
End Property

'FIXIT: Declare 'NewData' with an early-bound data type                                    FixIT90210ae-R1672-R1B8ZE
Private Sub AGIResource_InsertData(NewData As Variant, Optional ByVal InsertPos As Long = -1&)
  On Error GoTo ErrHandler
  
  agRes.InsertData NewData, InsertPos
Exit Sub

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Sub


Private Property Get AGIResource_Loaded() As Boolean
  AGIResource_Loaded = agRes.Loaded
End Property



Private Property Get AGIResource_Loc() As Long
  AGIResource_Loc = agRes.Loc

End Property


Private Sub AGIResource_NewResource(Optional ByVal Reset As Boolean = False)
  On Error GoTo ErrHandler
  
  agRes.NewResource Reset
Exit Sub

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Sub

Private Property Get AGIResource_Number() As Byte
  AGIResource_Number = agRes.Number
End Property


Private Function AGIResource_ReadByte(Optional ByVal lngPos As Long = MAX_RES_SIZE + 1) As Byte
  On Error GoTo ErrHandler
  
  AGIResource_ReadByte = agRes.ReadByte(lngPos)
Exit Function

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Function



Private Function AGIResource_ReadWord(Optional ByVal lngPos As Long = -1&, Optional ByVal blnMSLS As Boolean = False) As Long
  On Error GoTo ErrHandler
  
  AGIResource_ReadWord = agRes.ReadWord(lngPos)
Exit Function

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Function


Private Sub AGIResource_RemoveData(ByVal RemovePos As Long, Optional ByVal RemoveCount As Long = 1)

  On Error GoTo ErrHandler
  
  agRes.RemoveData RemovePos, RemoveCount
Exit Sub

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Sub

Private Property Get AGIResource_ResFile() As String
  AGIResource_ResFile = agRes.ResFile
End Property



Private Property Get AGIResource_ResType() As AGIResType
  AGIResource_ResType = agRes.ResType
End Property


Private Sub AGIResource_SetData(NewData() As Byte)

  agRes.SetData NewData
End Sub

Private Sub AGIResource_SetPos(ByVal lngPos As Long)
  On Error GoTo ErrHandler
  
  agRes.SetPos lngPos
Exit Sub

ErrHandler:
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Sub



Private Property Get AGIResource_Size() As Long
  AGIResource_Size = agRes.Size
End Property



Private Property Get AGIResource_SizeInVOL() As Long
  AGIResource_SizeInVOL = agRes.SizeInVol

End Property


Private Property Get AGIResource_Volume() As Long
  AGIResource_Volume = agRes.Volume
End Property



Private Sub agRes_Change()

  'reset CRC since compiled data and source code no longer match
  mCompiledCRC = 0
End Sub

Private Sub Class_Initialize()

  strErrSource = "WinAGI.Logic"
  Set agRes = New AGIResource
  agRes.SetType rtLogic
  
  'set default resource data
  agRes.NewResource
  agRes.WriteByte &H1, 0
  agRes.WriteByte &H0
  agRes.WriteByte &H0
  agRes.WriteByte &H0
  agRes.WriteByte &H2
  agRes.WriteByte &H0
  
  'set default source code
  Set mSourceCode = New SourceCode
  mSourceCode.Syntax = agMainLogSettings.DefaultSyntax
  If agMainLogSettings.DefaultSyntax = lsAGI Then
    mSourceCode.Text = "return();" & vbNewLine & vbNewLine & "//messages"
  Else
    mSourceCode.Text = "return()" & vbNewLine & vbNewLine & "'Messages"
  End If
  mSourceCode.IsDirty = True
  
  mIsRoom = True
  mResID = "NewLogic"
End Sub
Private Sub Class_Terminate()
  
  Dim i As Long
  
  Set agRes = Nothing
  Set mSourceCode = Nothing
  
End Sub

Private Sub mSourceCode_Change()
  
  'note that change has occured
End Sub


Private Sub mSourceCode_Loaded(LoadReason As Long)
  
  Select Case LoadReason
  Case 1 'loaded via decompile
    'if in a game
    If agRes.InGame Then
      'set compiled CRC (since we know logic IS compiled)
      mCompiledCRC = mSourceCode.CRC
      'store source crc
      WriteProperty PC_LOGIC, PT_CRC32, agRes.Number, Hex$(mCompiledCRC)
    End If
  
  Case 0 'loaded from file
  
  End Select
End Sub


