Attribute VB_Name = "LogCompile"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E24FD1F01B4"
  Option Explicit
  'set compare option to text since the compiler
  'is not case sensitive
  Option Compare Text

  Private tmpLogRes As AGIResource
  Private strAnd As String, strOr As String, strNot As String
  Private strEqual As String, strNotEqual As String
  Private strConst As String, strConstEqual As String
  Private strMsgMarker As String, strMsgEqual As String
  Private strCommentStart As String, strCommentEnd As String
  Private strCommentChar1 As String, strCommentChar2 As String
    
  Private mbytData() As Byte
  
  Private Type LogicGoto
    LabelNum  As Byte
    DataLoc As Integer
  End Type
  
  Private Const MaxBlockDepth = 64
  Private Const MaxLabels = 255
  Private Const UseTypeChecking = True
  Private Const NULL_MSG = ""
  
  Private Type LogicLabel
    Name  As String
    Loc As Integer
  End Type
  
  Private bytLogComp As Byte
  Private blnError As Boolean
  Private strErrMsg As String
  Private strModule As String
  Private lngErrLine As Long
  Private intCtlCount As Long
  Private strWarning As String
  Private blnWarnings As Boolean
  
  Private strIncludeFile() As String
  Private lngIncludeOffset As Long 'to correct line number due to added include lines
  Private stlInput As StringList   'the entire text to be compiled; includes the
  'original logic text, includes, and defines
  
  Private lngLine As Long
  Private lngPos As Long
  Private strCurrentLine As String
  Private blnContinueLine As Boolean  'true if a line is marked for continuation (VB only)
  
  Private strMsg(255) As String
  Private blnMsg(255) As Boolean
  
  Private llLabel(MaxLabels) As LogicLabel
  Private bytLabelCount As Byte
  
'  Private tdDefines(MaxDefines) As TDefine
  Private tdDefines() As TDefine
  Private lngDefineCount As Long

  Public blnSetIDs As Boolean
  Private strLogID() As String
  Private strPicID() As String
  Private strSndID() As String
  Private strViewID() As String
Private Sub CheckResFlagUse(ByVal ArgVal As Byte)

  'if error level is low, don't do anything
  If agMainLogSettings.ErrorLevel = leLow Then
    Exit Sub
  End If
  
  Select Case ArgVal
  Case 4, 7, 9, 10, Is >= 13
    'f4 = input_parsed
    'f7 = script_buffer_blocked
    'f9 = sound_on
    'f10 = trace_abled
    'f13 = inventory_select_enabled
    'f14 = menu_enabled
    'f15 = windows_remain
    'f16 = auto_restart
    'no restrictions
    
  Case Else 'all other reserved variables should be read only
    AddWarning Replace(LoadResString(5025), ARG1, agResFlag(ArgVal).Name)
  End Select
End Sub
Private Sub CheckResVarUse(ByVal ArgNum As Byte, ByVal ArgVal As Byte)

  'if error level is low, don't do anything
  If agMainLogSettings.ErrorLevel = leLow Then
    Exit Sub
  End If
  
  Select Case ArgNum
  Case Is >= 27, 21, 7, 3 'score, max score, msgbox delay time, and all non-restricted variables
    'no restrictions
  
  Case 6 'ego direction
    'should be restricted to values 0-8
    If ArgVal > 8 Then
      AddWarning Replace(Replace(LoadResString(5018), ARG1, agResVar(6).Name), ARG2, "8")
    End If
    
  Case 10 'cycle delay time
    'large values highly unusual
    If ArgVal > 20 Then
      AddWarning Replace(LoadResString(5055), ARG1, agResVar(10).Name)
    End If
    
  Case 23 'sound attenuation
    'restrict to 0-15
    If ArgVal > 15 Then
      AddWarning Replace(Replace(LoadResString(5018), ARG1, agResVar(23).Name), ARG2, "15")
    End If
    
  Case 24 'max input length
    If ArgVal > 39 Then
      AddWarning Replace(Replace(LoadResString(5018), ARG1, agResVar(24).Name), ARG2, "39")
    End If
    
  Case Else 'all other reserved variables should be read only
    AddWarning Replace(LoadResString(5017), ARG1, agResVar(ArgNum).Name)
  End Select
End Sub


Public Sub CompileLogic(SourceLogic As AGILogic)
  'this function compiles the sourcecode that is passed
  'the function returns a value of true if successful; it returns false
  'and sets information about the error if an error in the source text is found
  
  'note that when errors are returned, line is adjusted because
  'editor rows(lines) start at '1', but the compiler starts at line '0'
  
  Dim blnCompiled As Boolean
  Dim stlSource As StringList
  Dim fso As FileSystemObject
  Dim strCRC As String
  
  Debug.Assert SourceLogic.SourceCode.Syntax <> lsUnknown
  
  'set error and warning info to success as default
  blnError = False
  lngErrLine = -1
  strErrMsg = vbNullString
  strModule = vbNullString
  blnWarnings = False
  strWarning = vbNullString
  intCtlCount = 0
  
  On Error Resume Next
  'initialize global defines
  Set fso = New FileSystemObject
  'get datemodified property
  strCRC = CStr(fso.GetFile(agGameDir & "globals.txt").DateLastModified)
  Set fso = Nothing
  If CRC32(StrConv(strCRC, vbFromUnicode)) <> agGlobalCRC Then
    GetGlobalDefines
  End If
  
  'if ids not set yet
  If Not blnSetIDs Then
    SetResourceIDs
  End If
  
  On Error GoTo ErrHandler
  
  'assign to source stringlist
  Set stlSource = New StringList
  stlSource.Assign SourceLogic.SourceCode.Source.AllLines
    
  bytLogComp = SourceLogic.Number
  
  'if current syntax doesn't match logic syntax
  If SourceLogic.SourceCode.Syntax <> agCurrentSyntax Then
    'set syntax strings
    SetSyntax SourceLogic.SourceCode.Syntax
  End If
  
  'if predefines aren't correct
  If agPreDefSyntax <> agCurrentSyntax Then
    AssignReservedDefines agCurrentSyntax
  End If
  
  'reset error info
  lngErrLine = -1
  strErrMsg = vbNullString
  strModule = vbNullString
  
  'add include files
  If Not AddIncludes(stlSource) Then
    'dereference objects
    Set stlInput = Nothing
    'return error
    On Error GoTo 0
    On Error GoTo 0: Err.Raise vbObjectError + 635, "LogCompile", CStr(lngErrLine + 1) & "|" & strModule & "|" & strErrMsg
    Exit Sub
  End If
  
  'remove any blank lines from end
  Do Until Len(stlInput(stlInput.Count - 1)) <> 0
    stlInput.Delete stlInput.Count - 1
  Loop
  
  'strip out all comments
  If Not RemoveComments() Then
    'dereference objects
    Set stlInput = Nothing
    'return error
    On Error GoTo 0
    On Error GoTo 0: Err.Raise vbObjectError + 635, "LogCompile", CStr(lngErrLine + 1) & "|" & strModule & "|" & strErrMsg
    Exit Sub
  End If
  
  'read labels
  If Not ReadLabels() Then
    'dereference objects
    Set stlInput = Nothing
    'return error
    On Error GoTo 0
    On Error GoTo 0: Err.Raise vbObjectError + 635, "LogCompile", CStr(lngErrLine + 1) & "|" & strModule & "|" & strErrMsg
    Exit Sub
  End If
  
  'enumerate and replace all the defines
  If Not ReadDefines() Then
    'dereference objects
    Set stlInput = Nothing
    'return error
    On Error GoTo 0
    On Error GoTo 0: Err.Raise vbObjectError + 635, "LogCompile", CStr(lngErrLine + 1) & "|" & strModule & "|" & strErrMsg
    Exit Sub
  End If
  
  'read predefined messages
  If Not ReadMsgs() Then
    'dereference objects
    Set stlInput = Nothing
    'return error
    On Error GoTo 0
    On Error GoTo 0: Err.Raise vbObjectError + 635, "LogCompile", CStr(lngErrLine + 1) & "|" & strModule & "|" & strErrMsg
    Exit Sub
  End If
  
  'assign temporary resource object
  Set tmpLogRes = New AGIResource
  tmpLogRes.NewResource
  
  'write a word as a place holder for offset to msg section start
  tmpLogRes.WriteWord 0, 0
  
  'if using VB syntax
  Select Case agCurrentSyntax
  Case lsVB
    'use VB compiler
    blnCompiled = CompileVB()
  Case lsAGI
    'use c-like compiler
    blnCompiled = CompileAGI()
  End Select
  
  'compile commands
  If Not blnCompiled Then
    'dereference objects
    tmpLogRes.Unload
    Set tmpLogRes = Nothing
    Set stlInput = Nothing
    'return error
    On Error GoTo 0: Err.Raise vbObjectError + 635, "LogCompile", CStr(lngErrLine + 1) & "|" & strModule & "|" & strErrMsg
    Exit Sub
  End If
  
  'write message section
  If Not WriteMsgs() Then
    'dereference objects
    tmpLogRes.Unload
    Set tmpLogRes = Nothing
    Set stlInput = Nothing
    'return error
    On Error GoTo 0
    On Error GoTo 0: Err.Raise vbObjectError + 635, "LogCompile", CStr(lngErrLine + 1) & "|" & strModule & "|" & strErrMsg
    Exit Sub
  End If
  
  'assign resource data
  SourceLogic.Resource.AllData = tmpLogRes.AllData
  
  'update crc
  SourceLogic.SetCRC
  
  'dereference objects
  tmpLogRes.Unload
  Set tmpLogRes = Nothing
  Set stlInput = Nothing
  
  'check for warnings
  If blnWarnings Then
    'return error
    On Error GoTo 0
    On Error GoTo 0: Err.Raise vbObjectError + 636, "LogCompile", strWarning
    Exit Sub
  End If
Exit Sub

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  'dereference objects
  tmpLogRes.Unload
  Set tmpLogRes = Nothing
  Set stlInput = Nothing
  'return error
  On Error GoTo 0
  On Error GoTo 0: Err.Raise vbObjectError + 634, "LogCompile", "-1||" & Replace(LoadResString(591), ARG1, CStr(lngError) & ":" & strError)
End Sub


Private Function CompileVB() As Boolean
  'main compiler function for VB input
  'steps through input one command at a time and converts it
  'to AGI logic code
  'Note that don't need to set blnError flag here;
  'an error will cause this function to return a value of false
  'which causes the compiler to display error info
    
  Const MaxGotos = 255

  Dim strLine As String
  Dim strNextCmd As String
  Dim strPrevCmd As String
  Dim strArg As String, bytArg(7) As Byte
  Dim i As Integer
  Dim intCmdNum As Integer
  Dim BlockStartDataLoc(MaxBlockDepth) As Integer
  Dim BlockDepth As Integer
  Dim BlockIsIf(MaxBlockDepth) As Boolean
  Dim BlockLength(MaxBlockDepth) As Integer
  Dim CurLabel As Integer
  Dim intLabelNum As Integer
  Dim Gotos(MaxGotos) As LogicGoto
  Dim NumGotos As Integer
  Dim GotoData As Long
  Dim CurGoto As Integer
  Dim blnLastCmdRtn As Boolean
  Dim lngReturnLine As Long
  
  On Error GoTo ErrHandler
  
  'initialize variables
  BlockDepth = 0
  NumGotos = 0
  
  'reset compiler
  ResetCompiler
  
  'get first command
  strNextCmd = NextCommand()
    
  'process commands in the input string list until finished
  Do Until lngLine = -1
    'if last line wasn't blank AND last command was not 'return'
    If strPrevCmd <> vbNullString And strPrevCmd <> "return" Then
      'reset last command flag
      blnLastCmdRtn = False
      lngReturnLine = 0
    End If
    'process the command
    Select Case strNextCmd
    Case "end if"
      'if no block currently open,
      If BlockDepth = 0 Then
        strErrMsg = LoadResString(4053)
        Exit Function
      End If
      'if last position in resource is two bytes from start of block
      If tmpLogRes.Size = BlockStartDataLoc(BlockDepth) + 2 Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          strErrMsg = LoadResString(4049)
          Exit Function
        Case leMedium
          'set warning
          AddWarning LoadResString(5001)
        'Case leLow
          'no action
        End Select
      End If
      'calculate and write block length
      BlockLength(BlockDepth) = tmpLogRes.Size - BlockStartDataLoc(BlockDepth) - 2
      tmpLogRes.WriteWord CLng(BlockLength(BlockDepth)), CLng(BlockStartDataLoc(BlockDepth))
      'remove block from stack
      BlockDepth = BlockDepth - 1
      
      'should be at end of line
      If LenB(NextCommand()) <> 0 Then
        strErrMsg = LoadResString(4054)
        Exit Function
      End If
      'get next line
      IncrementLine
    Case "if"
      'compile the 'if' statement
      If Not CompileIf() Then
        Exit Function
      End If
      'if block stack exceeded
      If BlockDepth > MaxBlockDepth Then
        strErrMsg = Replace(LoadResString(4110), ARG1, CStr(MaxBlockDepth))
        Exit Function
      End If
      'add block to stack
      BlockDepth = BlockDepth + 1
      BlockStartDataLoc(BlockDepth) = tmpLogRes.GetPos
      BlockIsIf(BlockDepth) = True
      'write placeholders for block length
      tmpLogRes.WriteWord &H0

      'should be at end of line
      If LenB(NextCommand()) <> 0 Then
        strErrMsg = LoadResString(4054)
        Exit Function
      End If
      'get next line
      IncrementLine
    
    Case "else"
      'else acts as a closing bracket,
      'but only if currently working on an 'if' block
      
      'if no block currently open,
      If BlockDepth = 0 Then
        strErrMsg = LoadResString(4052)
        Exit Function
      End If
      
      'if current block is not an if
      If Not BlockIsIf(BlockDepth) Then
        strErrMsg = LoadResString(4052)
        Exit Function
      End If
      
      'if last position in resource is two bytes from start of block
      If tmpLogRes.Size = BlockStartDataLoc(BlockDepth) + 2 Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          strErrMsg = LoadResString(4049)
          Exit Function
        Case leMedium
          'set warning
          AddWarning LoadResString(5001)
        Case leLow
          'no action
        End Select
      End If
      
      'close out the if block
      'calculate and write block length (include 3 bytes to accomodate 'else')
      BlockLength(BlockDepth) = tmpLogRes.Size - BlockStartDataLoc(BlockDepth) - 2 + 3
      tmpLogRes.WriteWord CLng(BlockLength(BlockDepth)), CLng(BlockStartDataLoc(BlockDepth))
      
      'change this block to an 'else' block
      BlockIsIf(BlockDepth) = False
      'write the 'else' code
      tmpLogRes.WriteByte &HFE
      BlockStartDataLoc(BlockDepth) = tmpLogRes.GetPos
      tmpLogRes.WriteWord &H0  ' block length filled in later.
      
      'should be at end of line
      If LenB(NextCommand()) <> 0 Then
        strErrMsg = LoadResString(4054)
        Exit Function
      End If
      'get next line
      IncrementLine
    
    Case "goto"
      'next command should be a label
      strArg = NextCommand()
       
      'if argument is NOT a valid label
      If LabelNum(strArg) = 0 Then
        'if the arg is a bracket,
        If strArg = "(" Then
          'no brackets for VB gotos
          strErrMsg = LoadResString(4117)
        Else
          strErrMsg = Replace(LoadResString(4074), ARG1, strArg)
        End If
        Exit Function
      End If
      'if too many gotos
      If NumGotos >= MaxGotos Then
        strErrMsg = Replace(LoadResString(4108), ARG1, CStr(MaxGotos))
      End If
      'save this goto info on goto stack
      NumGotos = NumGotos + 1
      Gotos(NumGotos).LabelNum = LabelNum(strArg)
      'write goto command byte
      tmpLogRes.WriteByte &HFE
      Gotos(NumGotos).DataLoc = tmpLogRes.GetPos
      'write placeholder for amount of offset
      tmpLogRes.WriteWord &H0
      
      'should be at end of line
      If LenB(NextCommand()) <> 0 Then
        strErrMsg = LoadResString(4054)
        Exit Function
      End If
      'get next line
      IncrementLine
  
    Case vbNullString 'blank line
      'move to next line
      IncrementLine
    
    Case Else
      'must be a label, command, or special syntax
      'if next character is a colon
      If Mid$(strCurrentLine, lngPos + 1, 1) = ":" Then
        'it's a label
        intLabelNum = LabelNum(strNextCmd)
        'if not a valid label
        If intLabelNum = 0 Then
          strErrMsg = LoadResString(4076)
          Exit Function
        End If
        'save position of label
        llLabel(intLabelNum).Loc = tmpLogRes.Size
        'read in next char to skip past the colon
        NextChar
        'should be at end of line
        If LenB(NextCommand()) <> 0 Then
          strErrMsg = LoadResString(4054)
          Exit Function
        End If
        'get next line
        IncrementLine
      Else
        'get number of command
        intCmdNum = CommandNum(False, strNextCmd)
        'if invalid version
        If intCmdNum = 254 Then
          'raise error
          strErrMsg = Replace(LoadResString(4065), ARG1, strNextCmd)
          Exit Function
        'if command not found,
        ElseIf intCmdNum = 255 Then
          'try to parse special syntax
          If CompileSpecial(strNextCmd) Then
            'check for error
            If blnError Then
              Exit Function
            End If
          Else
            'unknown command
            strErrMsg = Replace(LoadResString(4116), ARG1, strNextCmd)
            Exit Function
          End If
        Else
          'write the command code,
          tmpLogRes.WriteByte CByte(intCmdNum)
          
          'next character MUST be "("
          If NextChar() <> "(" Then
            strErrMsg = LoadResString(4048)
            Exit Function
          End If
          
          'now extract arguments,
          For i = 0 To agCmds(CByte(intCmdNum)).ArgCount - 1
            'after first argument, verify comma separates arguments
            If i > 0 Then
              If NextChar(True) <> "," Then
                strErrMsg = Replace(LoadResString(4047), ARG1, CStr(i))
                Exit Function
              End If
            End If
            bytArg(i) = GetNextArg(agCmds(CByte(intCmdNum)).ArgType(i), i)
            'if error
            If blnError Then
              Exit Function
            End If
                        
            'write argument
            tmpLogRes.WriteByte bytArg(i)
          Next i
          
          'validate arguments for this command
          If Not ValidateArgs(intCmdNum, bytArg()) Then
            'error assigned by called function
            Exit Function
          End If
          
          'next character MUST be ")"
          If NextChar() <> ")" Then
            strErrMsg = LoadResString(4048)
            Exit Function
          End If
          
          If intCmdNum = 0 Then
            blnLastCmdRtn = True
            'save last line
            If lngReturnLine = 0 Then
              lngReturnLine = lngLine
            End If
          End If
        End If
        
        'should be at end of line
        If LenB(NextCommand()) <> 0 Then
          strErrMsg = LoadResString(4054)
          Exit Function
        End If
        'get next line
        IncrementLine
      End If
     End Select
     
    'if last command was NOT blank
    If LenB(strNextCmd) <> 0 Then
      'only track non-blank commands
      strPrevCmd = strNextCmd
    End If
    
    'get next command
    strNextCmd = NextCommand()
  Loop
  
  'if last command was not a return,
  If (Not blnLastCmdRtn) Then
    strErrMsg = LoadResString(4102)
    Exit Function
  End If
  
  'check to see if no blocks remain
  If BlockDepth > 0 Then
    strErrMsg = LoadResString(4043)
    'reset errorline position of return command
    lngErrLine = lngReturnLine
    Exit Function
  End If
  
  'write in goto values
  For CurGoto = 1 To NumGotos
    GotoData = llLabel(Gotos(CurGoto).LabelNum).Loc - Gotos(CurGoto).DataLoc - 2
    If GotoData < 0 Then
      'need to convert it to an unsigned integer value
      GotoData = &H10000 + GotoData
    End If
    tmpLogRes.WriteWord CLng(GotoData), CLng(Gotos(CurGoto).DataLoc)
  Next CurGoto
  
  'return true
  CompileVB = True
Exit Function
ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 655, strErrSrc, Replace(LoadResString(655), ARG1, CStr(lngError) & ":" & strError)
End Function


Public Function ConvertArgument(ByRef strArgIn As String, ByVal ArgType As ArgTypeEnum, Optional ByRef blnVarOrNum As Boolean = False) As Boolean
  'if input is not a system argument already
  '(i.e. ##, v##, f##, s##, o##, w##, i##, c##)
  'this function searches resource IDs, local defines, global defines,
  'and reserved names for strArgIn; if found
  'strArgIn is replaced with the value of the define
  'optional argtype is used to identify words, messages, and inv objects
  'to speed up search
  
  'NOTE: this does NOT validate the numerical value of arguments;
  'calling function is responsible to make that check
  'it also does not concatenate strings
  
  'to support calls from special syntax compilers, need to be able
  'to check for numbers AND variables with one check
  'the blnVarOrNum flag is used to do this; when the flag is
  'true, number searches also return variables
  
  Dim i As Long
  Dim intAsc As Integer
  
  On Error GoTo ErrHandler
  
  'check if already in correct format
  Select Case ArgType
  Case atNum  'numeric only
    If IsNumeric(strArgIn) Then
      ConvertArgument = True
      'reset VarOrNum flag
      blnVarOrNum = False
      Exit Function
    End If
    'unless looking for var or num
    If blnVarOrNum Then
      'then 'v##' is ok
      If (AscW(strArgIn) Or 32) = 118 Then
        If VariableValue(strArgIn) <> -1 Then
          ConvertArgument = True
          Exit Function
        End If
      End If
    End If
  
  Case atVar
    'if first char matches
    If (AscW(strArgIn) Or 32) = 118 Then
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    End If
    
  Case atFlag
    'if first char matches
    If (AscW(strArgIn) Or 32) = 102 Then
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    End If
    
  Case atCtrl
    'if first char matches
    If (AscW(strArgIn) Or 32) = 99 Then
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    End If
    
  Case atSObj
    'if first char matches
    If (AscW(strArgIn) Or 32) = 111 Then
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    End If
    
  Case atStr
    'if first char matches
    If (AscW(strArgIn) Or 32) = 115 Then
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    End If
    
  Case atWord 'NOTE: this is NOT vocab word; this is word arg type (used in command word.to.string)
    'if first char matches
    If (AscW(strArgIn) Or 32) = 119 Then
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    End If
    
  Case atMsg
    'if first char matches, or is a quote
    intAsc = AscW(strArgIn)
    Select Case intAsc
    Case 77, 109
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    Case 34
      'strings are always ok
      ConvertArgument = True
      Exit Function
    End Select
    
  Case atIObj
    'if first char matches, or is a quote
    intAsc = AscW(strArgIn)
    Select Case intAsc
    Case 73, 105
      'if this arg returns a valid value
      If VariableValue(strArgIn) <> -1 Then
        'ok
        ConvertArgument = True
        Exit Function
      End If
    Case 34
      'strings are always ok
      ConvertArgument = True
      Exit Function
    End Select
    
  Case atVocWrd
    'can be number or string in quotes
    If IsNumeric(strArgIn) Or AscW(strArgIn) = 34 Then
      'ok
      ConvertArgument = True
      Exit Function
    End If
  End Select
  
  'arg is not in correct format; must be reserved name, global or local define, or an error
  
  'first, check against local defines
  For i = 0 To lngDefineCount - 1
    If strArgIn = tdDefines(i).Name Then
      'match found; check that value is correct type
      Select Case ArgType
      Case atNum
        'check for number
        If IsNumeric(tdDefines(i).Value) Then
          'reset VarOrNum flag
          blnVarOrNum = False
          ConvertArgument = True
          strArgIn = tdDefines(i).Value
          Exit Function
        End If
        
        'if checking for variables
        If blnVarOrNum Then
          If (AscW(tdDefines(i).Value) Or 32) = 118 Then
            'if this define returns a valid value
            If VariableValue(tdDefines(i).Value) <> -1 Then
              'ok
              strArgIn = tdDefines(i).Value
              ConvertArgument = True
            End If
          End If
        End If
        
      Case atVar
        'v## only
        If (AscW(tdDefines(i).Value) Or 32) = 118 Then
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        End If
        
      Case atFlag
        'f## only
        If (AscW(tdDefines(i).Value) Or 32) = 102 Then
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        End If
        
      Case atMsg
        'm## or a string
        intAsc = AscW(tdDefines(i).Value)
        Select Case intAsc
        Case 77, 109
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        
        Case 34
          strArgIn = tdDefines(i).Value
          ConvertArgument = True
        End Select
        
      Case atSObj
        'o## only
        If AscW(tdDefines(i).Value) = 111 Then
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        End If
        
      Case atIObj
        'i## or a string
        intAsc = AscW(tdDefines(i).Value)
        Select Case intAsc
        Case 73, 105
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        Case 34
          strArgIn = tdDefines(i).Value
          ConvertArgument = True
        End Select
        
      Case atStr
        's## only
        If (AscW(tdDefines(i).Value) Or 32) = 115 Then
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        End If
        
      Case atWord
        'w## only
        If (AscW(tdDefines(i).Value) Or 32) = 119 Then
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        End If
        
      Case atCtrl
        'c## only
        If (AscW(tdDefines(i).Value) Or 32) = 99 Then
          'if this define returns a valid value
          If VariableValue(tdDefines(i).Value) <> -1 Then
            'ok
            strArgIn = tdDefines(i).Value
            ConvertArgument = True
          End If
        End If
        
      Case atVocWrd
        'numeric or string only
        If IsNumeric(tdDefines(i).Value) Then
          strArgIn = tdDefines(i).Value
          ConvertArgument = True
        ElseIf AscW(tdDefines(i).Value) = 34 Then
          strArgIn = tdDefines(i).Value
          ConvertArgument = True
        End If
        
      Case atDefStr
        'call to ConvertArgument is never made with type of atDefStr
      End Select
      'exit, regardless of result
      Exit Function
    End If
  Next i
  
  'second, check against global defines
  'for any type except vocab words
  If ArgType <> atVocWrd Then
    'check against this type of global defines
    For i = 0 To agGlobalCount - 1
      If agGlobal(i).Type = ArgType Then
        If strArgIn = agGlobal(i).Name Then
          strArgIn = agGlobal(i).Value
          'reset VarOrNum flag
          blnVarOrNum = False
          ConvertArgument = True
          Exit Function
        End If
      End If
    Next i
    'if checking var or num
    If blnVarOrNum Then
      'numbers were checked; need to check variables
      For i = 0 To agGlobalCount - 1
        If agGlobal(i).Type = atVar Then
          If strArgIn = agGlobal(i).Name Then
            strArgIn = agGlobal(i).Value
            ConvertArgument = True
            Exit Function
          End If
        End If
      Next i
    End If
  Else
    'check vocab words only against numbers
    For i = 0 To agGlobalCount - 1
      If agGlobal(i).Type = atNum Then
        If strArgIn = agGlobal(i).Name Then
          strArgIn = agGlobal(i).Value
          ConvertArgument = True
          Exit Function
        End If
      End If
    Next i
  End If
  
  'check messages, iobjs, and vocab words against global strings
  If (ArgType = atMsg) Or (ArgType = atIObj) Or (ArgType = atVocWrd) Then
    'check against global defines (string type)
    For i = 0 To agGlobalCount - 1
      If agGlobal(i).Type = atDefStr Then
        If strArgIn = agGlobal(i).Name Then
          strArgIn = agGlobal(i).Value
          ConvertArgument = True
          Exit Function
        End If
      End If
    Next i
  End If
  
  'third, check numbers against list of resource IDs
  If ArgType = atNum Then
    'check against resource IDs
    For i = 0 To 255
      'if this arg matches one of the resource ids
      If strArgIn = strLogID(i) Then
        strArgIn = CStr(i)
        'reset VarOrNum flag
        blnVarOrNum = False
        ConvertArgument = True
        Exit Function
      End If
      If strArgIn = strPicID(i) Then
        strArgIn = CStr(i)
        'reset VarOrNum flag
        blnVarOrNum = False
        ConvertArgument = True
        Exit Function
      End If
      If strArgIn = strSndID(i) Then
        strArgIn = CStr(i)
        'reset VarOrNum flag
        blnVarOrNum = False
        ConvertArgument = True
        Exit Function
      End If
      If strArgIn = strViewID(i) Then
        strArgIn = CStr(i)
        'reset VarOrNum flag
        blnVarOrNum = False
        ConvertArgument = True
        Exit Function
      End If
    Next i
  End If
  
  'lastly, if using reserved names,
  If agUseRes Then
    'last of all, check reserved names
    Select Case ArgType
    Case atNum
      For i = 0 To 4
        If strArgIn = agEdgeCodes(i).Name Then
          strArgIn = CStr(i)
          'reset VarOrNum flag
          blnVarOrNum = False
          ConvertArgument = True
          Exit Function
        End If
      Next i
      For i = 0 To 8
        If strArgIn = agEgoDir(i).Name Then
          strArgIn = CStr(i)
          'reset VarOrNum flag
          blnVarOrNum = False
          ConvertArgument = True
          Exit Function
        End If
      Next i
      For i = 0 To 4
        If strArgIn = agVideoMode(i).Name Then
          strArgIn = CStr(i)
          'reset VarOrNum flag
          blnVarOrNum = False
          ConvertArgument = True
          Exit Function
        End If
      Next i
      'check against invobj count
      If strArgIn = agResDef(4).Name Then
        strArgIn = CStr(agResDef(4).Value)
        'reset VarOrNum flag
        blnVarOrNum = False
        ConvertArgument = True
        Exit Function
      End If
      
      'if looking for numbers OR variables
      If blnVarOrNum Then
        'check against builtin variables as well
        For i = 0 To 26
          If strArgIn = agResVar(i).Name Then
            strArgIn = agResVar(i).Value
            ConvertArgument = True
            Exit Function
          End If
        Next i
      End If
      
    Case atVar
       For i = 0 To 26
        If strArgIn = agResVar(i).Name Then
          strArgIn = agResVar(i).Value
          ConvertArgument = True
          Exit Function
        End If
      Next i
      
    Case atFlag
       For i = 0 To 17
        If strArgIn = agResFlag(i).Name Then
          strArgIn = agResFlag(i).Value
          ConvertArgument = True
          Exit Function
        End If
      Next i
    Case atMsg
      For i = 1 To 3 'for gamever and gameabout and gameid
        If strArgIn = agResDef(i).Name Then
          strArgIn = agResDef(i).Value
          ConvertArgument = True
          Exit Function
        End If
      Next i
    Case atSObj
      If strArgIn = agResDef(0).Name Then
        strArgIn = agResDef(0).Value
        ConvertArgument = True
        Exit Function
      End If
    Case atStr
      If strArgIn = agResDef(5).Name Then
        strArgIn = agResDef(5).Value
        ConvertArgument = True
        Exit Function
      End If
    End Select
  End If
  
  'if not found, return false
ErrHandler:

  'just exit
End Function
Public Function ConvertToAGI(stlInput As StringList) As StringList
  'this function will take the input passed as stlInput
  'and attempt to convert from VB to C syntax
  
  Dim i As Long
  Dim strNewLine As String
  Dim strChar As String
  Dim blnInIf As Boolean
  Dim blnDefine As Boolean, blnSkipBlanks As Boolean
  Dim blnComment As Boolean, blnGoto As Boolean
  Dim intEndOfLine As Integer '0 = don't add ';' unless command found later
                              '1 = add ';', unless non-command found later
                              '2 = don't add ';' no matter what
  Dim intMsg As Integer '0 = not parsing msg
                        '1 = expect '('
                        '2 = expect ')'
                        '3 = expect '='
  Dim intCmd As Integer '0 = not parsing cmd
                        '1 = expect '('
                        '2 = expect ')'
                
  
  'if error encountered, just press on
  On Error Resume Next
  
  'if nothing passed,
  If stlInput Is Nothing Then
    'return nothing
    Set ConvertToAGI = Nothing
    Exit Function
  End If
  
  'if a game is not loaded, check to see if commands
  'have been assigned
  If Not agGameLoaded Then
    If LenB(agCmds(0).Name) = 0 Then
      AssignCommands
    End If
  End If
  
  'initialize output text
  Set ConvertToAGI = New StringList
  
  'if no text passed,
  If stlInput.Count = 0 Then
    Exit Function
  End If
  
  'set input format to VB
  SetSyntax lsVB
  
  'initialize compiler to first character of first line
  lngPos = 1
  lngLine = 0
  strCurrentLine = stlInput(0)
    
  Do Until lngLine = stlInput.Count
    'get next line
    strCurrentLine = stlInput(lngLine)
    'reset line variables
    strNewLine = vbNullString
    lngPos = 1
    intEndOfLine = 0
    blnComment = False
    blnSkipBlanks = False
    blnDefine = False
    intMsg = 0
    
    'check for continuation before processing line
    If Right$(strCurrentLine, 2) = " _" And Len(strCurrentLine) >= 2 Then
      'strip off continuation characters
      strCurrentLine = RTrim$(Left$(strCurrentLine, Len(strCurrentLine) - 2))
      'set end of line flag to indicate no ';'
      intEndOfLine = 2
    Else
      blnContinueLine = False
    End If
    Do Until lngPos > Len(strCurrentLine)
      'get next character
      strChar = Mid$(strCurrentLine, lngPos, 1)
      'if not space or tab
      If strChar <> " " And strChar <> vbTab Then
        'if expecting a parenthesis and this is NOT
        If intCmd = 1 And strChar <> "(" Then
          'add it (trimming off any preceding spaces)
          If LenB(RTrim$(strNewLine)) <> 0 Then
            strNewLine = RTrim$(strNewLine)
          End If
          strNewLine = strNewLine & "("
          intCmd = 2
        End If
      
        'reset skip blank flag
        blnSkipBlanks = False
        'get command that starts at lngpos (back up one so nextcommand works properly)
        lngPos = lngPos - 1
        strChar = NextCommand()
        'if adding a goto
        If blnGoto Then
          'add closting paragraph
          strChar = strChar & ")"
          blnGoto = False
        End If
        
        'if a text string,
        If Left$(strChar, 1) = QUOTECHAR Then
          'concatenate it
          strChar = ConcatArg(strChar)
          'Note: normally, strings will be concatenated across multiple lines in
          'VB syntax; however, because the continuation string is not stripped in
          'this function like it is in 'IncrementLine', the effect is that strings
          'across multiple lines in VB syntax are converted to c++ syntax multiple
          'string lines correctly
        End If
        
        Select Case strChar
        Case "const" 'check for const declaration
          'replace with #define
          strChar = "#define "
          'set define flag
          blnDefine = True
          'set skip blanksflag
          blnSkipBlanks = True
          'set end of line flag to indicate no ';'
          intEndOfLine = 2
        
        Case "=" 'check for define equal, or if comparison
          If blnDefine Then
            'add space
            strNewLine = RTrim$(strNewLine) & " "
            strChar = " "
            'skip blanks
            blnSkipBlanks = True
            'reset define flag
            blnDefine = False
          ElseIf intMsg = 3 Then
            strChar = vbNullString
            'add trailing space to current line
            strNewLine = strNewLine & " "
            'skip any extra spaces
            blnSkipBlanks = True
            'reset message parsing flag
            intMsg = 0
          ElseIf blnInIf Then
            'replace with dbl equal
            strChar = "=="
          End If
        Case Chr$(39), "rem" 'if comment found
          'if char is 'rem'
          If strChar = "rem" Then
            'increment pointer to skip space after 'rem'
            lngPos = lngPos + 1
          End If
          'set flag to comment out rest of line
          blnComment = True
        Case "_"  'check for continuation
          If lngPos > 1 Then
            If AscW(Mid$(strCurrentLine, lngPos - 1)) = 32 Then 'space
              'set end of line flag to indicate no ';'
              intEndOfLine = 2
              'skip character
              strChar = vbNullString
            End If
          End If
          
        Case "then"
          'convert
          strChar = ") {"
          'trim off spaces from current line
          strNewLine = RTrim$(strNewLine)
          'set if flag to false
          blnInIf = False
          'set end of line flag to indicate no ';'
          intEndOfLine = 0
        Case "if"
          'add paranthesis
          strChar = "if("
          blnSkipBlanks = True
          'set if flag
          blnInIf = True
          'set end of line flag to indicate no ';'
          'intEndOfLine = 0
        Case "else"
          'add closing bracket
          ConvertToAGI.Add strNewLine & "}"
          'indent new line same number of spaces
          strNewLine = Space$(Len(strNewLine))
          'convert
          strChar = "else {"
          'set end of line flag to indicate no ';'
          intEndOfLine = 2
        Case "end if"
          'convert
          strChar = "}"
          'set end of line flag to indicate no ';'
          intEndOfLine = 2
        Case "not"
          If blnInIf Then
            strChar = "!"
            blnSkipBlanks = True
          End If
        Case "<>", "><"
          If blnInIf Then
            strChar = "!="
          End If
        Case "and"
          If blnInIf Then
            strChar = "&& "
            blnSkipBlanks = True
          End If
        Case "or"
          If blnInIf Then
            strChar = "|| "
            blnSkipBlanks = True
          End If
        Case ":"
          'label
          intEndOfLine = 0
        Case "goto"
          'convert
          strChar = "goto("
          blnSkipBlanks = True
          blnGoto = True
        Case "message"
          'convert
          strChar = "#message "
          'set flag to indicate '(' is expected next
          intMsg = 1
        Case "("
          'if expecting '(' for msg
          If intMsg = 1 Then
            'strip out '('
            strChar = vbNullString
            'skip spaces
            blnSkipBlanks = True
            'set flag to indicate ')' expected
            '(presumably after a number)
            intMsg = 2
          End If
          'if expecting '(' for cmd
          If intCmd = 1 Then
            intCmd = 2
          End If
        Case ")"
          'if expecting ')' for msg
          If intMsg = 2 Then
            'srip out ')'
            strChar = vbNullString
            'skip any extra spaces
            blnSkipBlanks = True
            'set flag to indicate '=' expected
            intMsg = 3
          End If
          'if expecting ')' for cmd
          If intCmd = 2 Then
            intCmd = 0
          End If
        
        Case "*"
          'if NOT in an if block, skip blanks
          blnSkipBlanks = Not blnInIf
          
        Case Else
          'if using defined variables
          If agUseRes Then
            'check alphabetic strings against reserved values and commands
            Select Case AscW(strChar)
            Case 65 To 90, 97 To 122
              Do
                'check for reserved variables, flags, contants
                For i = 0 To 26
                  If strChar = agResVar(i).Name Then
                    strChar = "rv|" & CStr(i) & "|"
                    Exit Do
                  End If
                Next i
                For i = 0 To 17
                  If strChar = agResFlag(i).Name Then
                    strChar = "rf|" & CStr(i) & "|"
                    Exit Do
                  End If
                Next i
                For i = 0 To 4
                  If strChar = agEdgeCodes(i).Name Then
                    strChar = "ec|" & CStr(i)
                    Exit Do
                  End If
                Next i
                For i = 0 To 8
                  If strChar = agEgoDir(i).Name Then
                    strChar = "ed|" & CStr(i)
                    Exit Do
                  End If
                Next i
                For i = 0 To 4
                  If strChar = agVideoMode(i).Name Then
                    strChar = "vm|" & CStr(i)
                    Exit Do
                  End If
                Next i
                For i = 0 To 5
                  If strChar = agResDef(i).Name Then
                    strChar = "rd|" & CStr(i)
                    Exit Do
                  End If
                Next i
              
                'check for command
                For i = 0 To 181
                  If strChar = agCmds(i).Name Then
                    'expecting an opening parenthesis
                    intCmd = 1
                    Exit Do
                  End If
                Next i
              Loop Until True
            End Select
          End If
          
          'if not a hard eol exception
          If intEndOfLine = 0 Then
            'set it to indicate end of line is needed
            intEndOfLine = 1
          End If
        End Select
      End If
      'if in a comment
      If blnComment Then
        'if not continuing
        If intEndOfLine = 1 And Right$(Trim$(strNewLine), 1) <> QUOTECHAR Then
          strNewLine = RTrim$(strNewLine) & "; "
        End If
        'add rest of line
        strNewLine = strNewLine & "//" & Right$(strCurrentLine, Len(strCurrentLine) - lngPos)
        lngPos = Len(strCurrentLine)
        'NEVER add semicolon after a comment
        intEndOfLine = 2
      'if not skipping blanks OR this is  a NOT a space
      ElseIf Not blnSkipBlanks Or (strChar <> " " And strChar <> vbTab) Then
        'add charcter to current line
        strNewLine = strNewLine & strChar
      End If
      
      'increment pointer
      lngPos = lngPos + 1
    Loop
    'if at end of line, and still expecting a ')'
    If intEndOfLine = 1 And intCmd = 2 Then
      'add it
      strNewLine = strNewLine & ")"
      intCmd = 0
    ElseIf intEndOfLine = 1 And intCmd = 1 Then
      'add both parentheses
      strNewLine = strNewLine & "()"
      intCmd = 0
    End If
    
    'if end of line is set (and NOT a quote mark)
    If intEndOfLine = 1 And Right$(Trim$(strNewLine), 1) <> QUOTECHAR Then
      strNewLine = RTrim$(strNewLine) & ";"
    End If
    'add line
    ConvertToAGI.Add strNewLine
    'get next line
    lngLine = lngLine + 1
  Loop
End Function
Public Function ConvertToVB(stlInput As StringList) As StringList
  'this function will take the input passed as stlInput
  'and attempt to convert from c++ to VB
  
  Dim i As Long
  Dim strNewLine As String
  Dim strChar As String, strPrevCmd As String
  Dim blnInCommentBlock As Boolean
  Dim intGoto As Integer  '0 = not parsing goto
                          '1 = expecting '('
                          '2 = expecting ')'
  Dim intInIf As Integer  '0 = not parsing if
                          '1 = expecting '('
                          '2 = expecting '}'
  Dim intElse As Integer  '0 = not parsing else
                          '1 = expecting '{'
                          '2 = expecting '}'
  Dim intMsg As Integer   '0 = not parsing message
                          '1 = expect msg number
                          
                          
  Dim blnDefine As Boolean, blnSkipBlanks As Boolean
  Dim lngEndIfPos As Long, lngEndIfLine As Long
  Dim strContinuation As String
  Dim strEndChar As String
  Dim blnComment As Boolean
  
  'if error encountered, just press on
  On Error Resume Next
    
  'if nothing passed,
  If stlInput Is Nothing Then
    'return nothing
    Set ConvertToVB = Nothing
    Exit Function
  End If
  
  'initialize output text
  Set ConvertToVB = New StringList
  
  'if no text passed,
  If stlInput.Count = 0 Then
    Exit Function
  End If
  
  'set format to c
  SetSyntax lsAGI
  
  'initialize compiler to first character of first line
  lngPos = 1
  lngLine = 0
    
  Do Until lngLine = stlInput.Count
    'get next line
    strCurrentLine = RTrim$(stlInput(lngLine))
    'get ending character for this line
    strEndChar = Right$(strCurrentLine, 1)
    'if not an 'end of line' character (or end of line)
    If Len(strEndChar) = 1 And strEndChar <> ";" And strEndChar <> "{" And strEndChar <> "}" And strEndChar <> ":" And lngLine <> stlInput.Count - 1 Then
      'if ending in a quote
      If strEndChar = QUOTECHAR Then
        'if next line starts with a string
        If Left$(LTrim$(stlInput(lngLine + 1)), 1) = QUOTECHAR Then
          'use concatenation operator
          strContinuation = " & _"
        Else
          strContinuation = vbNullString
        End If
      Else
        'use VB continuation character
        strContinuation = " _"
      End If
    Else
      'this line ends a command
      strContinuation = vbNullString
    End If
    If blnInCommentBlock Then
      strContinuation = vbNullString
    End If
    'reset line variables
    strNewLine = vbNullString
    lngPos = 1
    blnComment = False
    blnSkipBlanks = False
    blnDefine = False
    
    Do Until lngPos > Len(strCurrentLine)
      'get next character
      strChar = Mid$(strCurrentLine, lngPos, 1)
      'if not space or tab
      If strChar <> Chr$(32) And strChar <> vbTab Then
        'if prev cmd NOT asterisk and not skipping blanks
        If strPrevCmd <> "*" And Not blnSkipBlanks Then
          'reset skip blank flag
          blnSkipBlanks = False
        End If
        'get command that starts at lngpos (back up one so nextcommand works properly)
        lngPos = lngPos - 1
        strChar = NextCommand()
        
        'if in a comment block
        If blnInCommentBlock Then
          'add comment to front
          If LenB(strNewLine) = 0 Then
            strNewLine = Chr$(39)
          End If
          
          'if command is an end block
          If strChar = "*" And Mid$(strCurrentLine, lngPos + 1, 1) = "/" Then
            lngPos = lngPos + 1
            
            'if NOT at end of line,
            If lngPos <> Len(strCurrentLine) Then
              'insert carriage return (by writing this line
              'and starting a new one)
              ConvertToVB.Add strNewLine
              strNewLine = vbNullString
            End If
            'dont add the asterisk (its part of the '*/' marker)
            strChar = vbNullString
            blnInCommentBlock = False
          End If
        Else
          'if previous cmd was message cmd
          If strPrevCmd = "Message(" Then
            'assume this cmd is the msg number
            strChar = strChar & ") = "
            blnSkipBlanks = True
          End If
          
          Select Case strChar
          Case ";"
            'end of line
            intGoto = 0
            intMsg = 0
            intInIf = 0
            intElse = 0
            'if not already at end of line
            If lngPos <> Len(strCurrentLine) Then
              'save current pos
              i = lngPos
              'if next command is NOT a comment
              strChar = NextCommand()
              If strChar <> "//" And strChar <> "[" And strChar <> "/*" Then
                'force new line
                ConvertToVB.Add strNewLine
                'use same indent
                strNewLine = Space$(Len(strNewLine) - Len(LTrim$(strNewLine)))
              End If
              'backup lngPos so next command is returned correctly
              lngPos = i
              'don't skip blanks after end of line
              blnSkipBlanks = False
            Else
              'end of likne; skip any blanks
              blnSkipBlanks = True
            End If
            'skip this character
            strChar = vbNullString
            
          Case "#define" 'check for define declaration
            'replace with const
            strChar = "Const "
            'get next cmd
            strChar = strChar & NextCommand()
            'if not at end?
            If lngPos <> Len(strCurrentLine) Then
              'add equal sign
              strChar = strChar & " = "
            End If
            
            'set skip blanksflag
            blnSkipBlanks = True
            strContinuation = vbNullString
            
          Case "/"
            If lngPos < Len(strCurrentLine) Then
              'need to check for comment start
              If AscW(Mid$(strCurrentLine, lngPos + 1)) = 42 Then '"*"
                lngPos = lngPos + 1
                'set flag to indicate in comment block
                blnInCommentBlock = True
                strContinuation = vbNullString
                'convert
                strChar = Chr$(39)
              End If
            End If
          Case "=="
            'replace with '='
            strChar = "="
          Case "//", "[" 'if comment found
            'set flag to comment out rest of line
            blnComment = True
            'if not at end of a line AND in an if statement
            If intGoto <> 0 Or intInIf <> 0 Or intElse <> 0 Or intMsg <> 0 Then
              'add continuation line
              strNewLine = strNewLine & strContinuation
            End If
            'reset continuation
            strContinuation = vbNullString
          Case "if"
            'add paranthesis
            strChar = "If "
            blnSkipBlanks = True
            'set if flag to indicate expecting '('
            intInIf = 1
          Case "("
            'if expecting '(' after an if
            If intInIf = 1 Then
              strChar = vbNullString
              blnSkipBlanks = True
              'set flag to indicate '{' expected
              intInIf = 2
            'if expecting '(' after goto
            ElseIf intGoto = 1 Then
              strChar = " "
              blnSkipBlanks = True
              'set flag to indicate ')' expected
              intGoto = 2
            End If
          Case ")"
            'if expecting ')' after goto
            If intGoto = 2 Then
              strChar = vbNullString
              intGoto = 0
            End If
          Case "else"
            'if immediately following a 'End If'
            If strPrevCmd = "End If" Then
              'get line with 'end if' that was written for the previous bracket
              strChar = ConvertToVB.StringLine(lngEndIfLine)
              ConvertToVB.StringLine(lngEndIfLine) = Left$(strChar, lngEndIfPos - 1) & Right$(strChar, Len(strChar) - 5 - lngEndIfPos)
              'if nothing left
              If LenB(Trim$(ConvertToVB.StringLine(lngEndIfLine))) = 0 Then
                ConvertToVB.Delete lngEndIfLine
              End If
            End If
            'convert
            strChar = "Else "
            'set flag to indicate expecting "{"
            intElse = 1
          Case "{"
            'if expecting bracket after if
            If intInIf = 2 Then
              'strip off trailing ')'
              If Right$(RTrim$(strNewLine), 1) = ")" Then
                strNewLine = Left$(RTrim$(strNewLine), Len(RTrim$(strNewLine)) - 1)
              End If
              'convert
              strChar = " Then"
              intInIf = 0
            ElseIf intElse = 1 Then
              'hide
              strChar = vbNullString
              intElse = 2
            End If
          Case "}"
            'always use 'end if'
            'if an 'else' is found, the 'end if'
            'will be deleted
            strChar = "End If"
            'save line and position of the bracket
            'incase it needs to be removed because
            'of an 'else' command
            lngEndIfLine = ConvertToVB.Count
            lngEndIfPos = lngPos
            
          Case "!"
            If intInIf = 2 Then
              strChar = "Not "
              blnSkipBlanks = True
            End If
          Case "!="
            If intInIf = 2 Then
              strChar = "<>"
            End If
          Case "&&"
            If intInIf = 2 Then
              strChar = " And "
              blnSkipBlanks = True
            End If
          Case "||"
            If intInIf = 2 Then
              strChar = " Or "
              blnSkipBlanks = True
            End If
          Case "goto"
            'convert
            strChar = "Goto "
            blnSkipBlanks = True
            intGoto = 1
          Case "#message"
            strChar = "Message("
            blnSkipBlanks = True
            intMsg = 1
          Case "++"
            'convert to long increment format
            strChar = " = " & strPrevCmd & " + 1"
          Case "--"
            'convert to long decrement format
            strChar = " = " & strPrevCmd & " - 1"
          Case "+="
            'convert to long math
            strChar = " = " & strPrevCmd & " + "
          Case "-="
            'convert to long math
            strChar = " = " & strPrevCmd & " - "
          Case "/="
            'convert to long math
            strChar = " = " & strPrevCmd & " / "
          Case "*="
             'convert to long math
            strChar = " = " & strPrevCmd & " * "
          Case "*"
            'if NOT in an if, skip blanks
            blnSkipBlanks = (intInIf = 0)
            
          Case Else
            'if using reserved variables
            If agUseRes Then
              'check alphabetic strings against reserved variables, flags, constants
              Do
                Select Case AscW(strChar)
                Case 65 To 90, 97 To 122
                  For i = 0 To 26
                    If strChar = agResVar(i).Name Then
                      strChar = "rv|" & CStr(i) & "|"
                      Exit Do
                    End If
                  Next i
                  For i = 0 To 17
                    If strChar = agResFlag(i).Name Then
                      strChar = "rf|" & CStr(i) & "|"
                      Exit Do
                    End If
                  Next i
                  For i = 0 To 4
                    If strChar = agEdgeCodes(i).Name Then
                      strChar = "ec|" & CStr(i)
                      Exit Do
                    End If
                  Next i
                  For i = 0 To 8
                    If strChar = agEgoDir(i).Name Then
                      strChar = "ed|" & CStr(i)
                      Exit Do
                    End If
                  Next i
                  For i = 0 To 4
                    If strChar = agVideoMode(i).Name Then
                      strChar = "vm|" & CStr(i)
                      Exit Do
                    End If
                  Next i
                  For i = 0 To 5
                    If strChar = agResDef(i).Name Then
                      strChar = "rd|" & CStr(i)
                      Exit Do
                    End If
                  Next i
                End Select
              Loop Until True
            End If
          End Select
          
          'save prev cmd
          strPrevCmd = strChar
        End If
      End If
      
      'if in a comment
      If blnComment Then
        'add rest of line
        strNewLine = strNewLine & Chr$(39) & Right$(strCurrentLine, Len(strCurrentLine) - lngPos)
        lngPos = Len(strCurrentLine)
      'if not skipping blanks OR this is  a NOT a space
      ElseIf Not blnSkipBlanks Or (strChar <> " " And strChar <> vbTab) Then
        'add charcter to current line
        strNewLine = strNewLine & strChar
      End If
      
      'increment pointer
      lngPos = lngPos + 1
    Loop
    'add line
    ConvertToVB.Add strNewLine & strContinuation
    'get next line
    lngLine = lngLine + 1
  Loop
End Function



Private Function GetNextArg(ByVal ArgType As ArgTypeEnum, ByVal ArgPos As Long, Optional ByRef blnVarOrNum As Boolean = False) As Long
  'this function retrieves the next argument and validates
  'that the argument is of the correct type
  'and has a valid value
  'multline message/string/inv.item/word strings are recombined, and checked
  'for validity
  'if successful, the function returns the value of the argument
  'if unsuccessful, the function sets the error flag and error msg
  '(in which case, the return value is meaningless)
  
  'special syntax compilers look for variables OR strings;
  'when argtype is atNum and this flag is set, numbers OR
  'variables return true; the flag is set to TRUE if returned value
  'is for a variable, to false if it is for a number
  
  Dim strArg As String, lngArg As Integer
  Dim i As Long
  
  On Error GoTo ErrHandler
  
'  'if using pos
'  If ArgPos > 0 Then
'    Cstr(ArgPos) = "for argument " & CStr(ArgPos)
'  End If
  
  'get next command
  strArg = NextCommand()
  
  'convert it
  If Not ConvertArgument(strArg, ArgType, blnVarOrNum) Then
    'error
    blnError = True
    strErrMsg = Replace(LoadResString(4063), ARG1, CStr(ArgPos))
    Exit Function
  End If
  
  Select Case ArgType
  Case atNum  'number
    'verify type is number
    If Not IsNumeric(strArg) Then
      'if NOT catching variables too
      If Not blnVarOrNum Then
        blnError = True
        strErrMsg = Replace(LoadResString(4062), ARG1, CStr(ArgPos))
        Exit Function
      End If
    Else
      'return 'is NOT a variable'; ensure flag is reset
      blnVarOrNum = False
    End If
    'convert to number and validate
    lngArg = VariableValue(strArg)
    If lngArg = -1 Then
      blnError = True
      strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
      Exit Function
    End If
  
  Case atVar, atFlag  'variable, flag
    'get value
    lngArg = VariableValue(strArg)
    If lngArg = -1 Then
      blnError = True
      strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
      Exit Function
    End If
  
  Case atCtrl    'controller
    'controllers should be  0 - 49
    'get value
    lngArg = VariableValue(strArg)
    If lngArg = -1 Then
      blnError = True
      'if high errlevel
      If agMainLogSettings.ErrorLevel = leHigh Then
        strErrMsg = Replace(LoadResString(4136), ARG1, CStr(ArgPos))
      Else
        strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
      End If
      Exit Function
    Else
      'if outside expected bounds (controllers should be limited to 0-49)
      If lngArg > 49 Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          'generate error
          blnError = True
          strErrMsg = Replace(LoadResString(4136), ARG1, CStr(ArgPos))
          Exit Function
          
        Case leMedium
          'generate warning
          AddWarning LoadResString(5060)
        End Select
      End If
    End If
  
  Case atSObj 'screen object
    'get value
    lngArg = VariableValue(strArg)
    If lngArg = -1 Then
      blnError = True
      strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
      Exit Function
    End If
    
    'check against max screen object value
    If lngArg > agInvObj.MaxScreenObjects Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        'generate error
        blnError = True
        strErrMsg = Replace(LoadResString(4119), ARG1, CStr(agInvObj.MaxScreenObjects))
        Exit Function
        
      Case leMedium
        'generate warning
        AddWarning Replace(LoadResString(5006), ARG1, CStr(agInvObj.MaxScreenObjects))
      End Select
    End If

  Case atStr 'string
    'get value
    lngArg = VariableValue(strArg)
    If lngArg = -1 Then
      blnError = True
      'if high errlevel
      If agMainLogSettings.ErrorLevel = leHigh Then
        'for version 2.089, 2.272, and 3.002.149 only 12 strings
        Select Case agIntVersion
        Case 2.089, 2.272, 3.002149
          strErrMsg = Replace(Replace(LoadResString(4079), ARG1, CStr(ArgPos)), ARG2, "11")
        Case Else
          strErrMsg = Replace(Replace(LoadResString(4079), ARG1, CStr(ArgPos)), ARG2, "23")
        End Select
        
      Else
        strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
      End If
      Exit Function
    Else
      'if outside expected bounds (strings should be limited to 0-23)
      If (lngArg > 23) Or (lngArg > 11 And (agIntVersion = 2.089 Or agIntVersion = 2.272 Or agIntVersion = 3.002149)) Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          'generate error
          blnError = True
          'for version 2.089, 2.272, and 3.002.149 only 12 strings
          Select Case agIntVersion
          Case 2.089, 2.272, 3.002149
            strErrMsg = Replace(Replace(LoadResString(4079), ARG1, CStr(ArgPos)), ARG2, "11")
          Case Else
            strErrMsg = Replace(Replace(LoadResString(4079), ARG1, CStr(ArgPos)), ARG2, "23")
          End Select
          Exit Function
          
        Case leMedium
         'generate warning
         'for version 2.089, 2.272, and 3.002.149 only 12 strings
          Select Case agIntVersion
          Case 2.089, 2.272, 3.002149
            AddWarning Replace(LoadResString(5007), ARG1, "11")
          Case Else
            AddWarning Replace(LoadResString(5007), ARG1, "23")
          End Select
        End Select
      End If
    End If
    
  Case atWord 'word  (word type is NOT words from word.tok)
    'get value
    lngArg = VariableValue(strArg)
    If lngArg = -1 Then
      blnError = True
      'if high error level
      If agMainLogSettings.ErrorLevel = leHigh Then
        strErrMsg = Replace(LoadResString(4090), ARG1, CStr(ArgPos))
      Else
        strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
      End If
      Exit Function
    Else
      'if outside expected bounds (words should be limited to 0-9)
      If lngArg > 9 Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          'generate error
          blnError = True
          strErrMsg = Replace(LoadResString(4090), ARG1, CStr(ArgPos))
          Exit Function
          
        Case leMedium
          'generate warning
          AddWarning LoadResString(5008)
        End Select
      End If
    End If
  
  Case atMsg  'message
    'returned arg is either m## or "msg"
    Select Case AscW(strArg)
    Case 109
      'validate value
      lngArg = VariableValue(strArg)
      If lngArg = -1 Then
        blnError = True
        strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
        Exit Function
        'verify msg exists
        If Not blnMsg(lngArg) Then
          strErrMsg = Replace(LoadResString(4113), ARG1, CStr(lngArg))
          Exit Function
        End If
      End If
    Case 34
      'concatenate, if applicable
      strArg = ConcatArg(strArg)
      If blnError Then
        'concatenation error; exit
        Exit Function
      End If
      
      'strip off quotes
      strArg = Mid$(strArg, 2, Len(strArg) - 2)
      'convert to msg number
      lngArg = MessageNum(strArg)
      'if valid number not found
      If lngArg = 0 Then
        blnError = True
        strErrMsg = LoadResString(4092)
        Exit Function
      End If
    End Select
    
  Case atIObj 'inventory object
      
    'if character is inv obj arg type prefix
    Select Case AscW(strArg)
    Case 105
      'validate value
      lngArg = VariableValue(strArg)
      If lngArg = -1 Then
        blnError = True
        strErrMsg = Replace(LoadResString(4066), ARG1, CStr(ArgPos))
        Exit Function
      End If
      
    Case 34
      'concatenate, if applicable
      strArg = ConcatArg(strArg)
      If blnError Then
        'concatenation error
        Exit Function
      End If
      
      'convert to inv obj number
      'first strip off starting and ending quotes
      strArg = Mid$(strArg, 2, Len(strArg) - 2)
      'if a quotation mark is part of an object name,
      'it is coded in the logic as a '\"' not just a '"'
      'need to ensure all '\"' codes are converted to '"'
      'otherwise the object would never match
      strArg = Replace(strArg, "\""", QUOTECHAR)
      
      'step through all object names
      For i = 1 To agResDef(4).Value
        'if this is the object
        If strArg = agInvObj(i).ItemName Then
          'return this value
          lngArg = CByte(i)
          Exit For
        End If
      Next i
      
      'if not found,
      If i = agResDef(4).Value + 1 Then
        blnError = True
        strErrMsg = Replace(LoadResString(4075), ARG1, CStr(ArgPos))
        Exit Function
      End If
      
      'if object is not unique
      If Not agInvObj(lngArg).Unique Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          blnError = True
          strErrMsg = Replace(LoadResString(4036), ARG1, CStr(ArgPos))
          Exit Function
        Case leMedium
          'set warning
          AddWarning Replace(LoadResString(5003), ARG1, CStr(ArgPos))
        Case leLow
          'no action
        End Select
      End If
    End Select
    
    'if object number exceeds current object count,
    'If lngArg >= agInvObj.Count Then
    If lngArg > agResDef(4).Value Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        blnError = True
        strErrMsg = Replace(LoadResString(4112), ARG1, CStr(ArgPos))
        Exit Function
      Case leMedium
        'set warning
        AddWarning Replace(LoadResString(5005), ARG1, CStr(ArgPos))
      Case leLow
        'no action
      End Select
    Else
      'if object is a question mark
      If agInvObj(lngArg).ItemName = "?" Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          blnError = True
          strErrMsg = Replace(LoadResString(4111), ARG1, CStr(ArgPos))
          Exit Function
        Case leMedium
          'set warning
          AddWarning LoadResString(5004)
        Case leLow
          'no action
        End Select
      End If
    End If
      
  Case atVocWrd
    'words can be ## or "word"
    If IsNumeric(strArg) Then
      lngArg = CLng(strArg)
    Else
      'this is a string; concatenate if applicable
      strArg = ConcatArg(strArg)
      If blnError Then
        'concatenation error
        Exit Function
      End If
      
      'convert to inv obj number
      'first strip off starting and ending quotes
      strArg = Mid$(strArg, 2, Len(strArg) - 2)
      
      On Error Resume Next
      'get argument val by checking against word list
      lngArg = agVocabWords(strArg).Group
      'if error
      If Err.Number <> 0 Then
        'set it to -1 so it will be trapped during validation
        lngArg = -1
      End If
    End If
    
    'use error trapping to validate word
    On Error Resume Next
    
    'verify this word number corresponds to a group
    lngArg = agVocabWords.GroupN(lngArg).GroupNum
    'if error
    If Err.Number <> 0 Then
      Err.Clear
      'if arg vlaue=-1 OR high level,
      If agMainLogSettings.ErrorLevel = leHigh Or (lngArg = -1) Then
        blnError = True
        strErrMsg = Replace(LoadResString(4114), ARG1, CStr(ArgPos))
        Exit Function
      Else
        If agMainLogSettings.ErrorLevel = leMedium Then
          'set warning
          AddWarning Replace(LoadResString(5019), ARG1, strArg)
        End If
      End If
    End If
  End Select
  
  'set return value
  GetNextArg = lngArg
Exit Function

ErrHandler:
  Debug.Assert False

  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "GetNextArg")
  blnError = True
End Function


Private Sub IncrementLine()
  'increments the the current line of input being processed
  'sets all counters, pointers, etc
  'as well as info needed to support error locating
  
  'if at end of input (lngLine=-1)
  If lngLine = -1 Then
    'just exit
    Exit Sub
  End If
  
  'if compiler is reset
  If lngLine = -2 Then
    'set it to -1 so line 0 is returned
    lngLine = -1
  End If
  
  'increment line counter
  lngLine = lngLine + 1
  lngPos = 0
  'if at end,
  If lngLine = stlInput.Count Then
    lngLine = -1
    Exit Sub
  End If
  'check for include lines
  If Left$(stlInput(lngLine), 2) = "#I" Then
    lngIncludeOffset = lngIncludeOffset + 1
    'set module
    strModule = strIncludeFile(CInt(Mid$(stlInput(lngLine), 3, InStr(3, stlInput(lngLine), ":") - 3)))
    'set errline
    lngErrLine = CLng(Mid$(stlInput(lngLine), InStr(3, stlInput(lngLine), ":") + 1, InStr(3, stlInput(lngLine), "#") - 5))
    strCurrentLine = Right$(stlInput(lngLine), Len(stlInput(lngLine)) - InStr(2, stlInput(lngLine), "#"))
  Else
    strModule = vbNullString
    lngErrLine = lngLine - lngIncludeOffset
    'set string
    strCurrentLine = stlInput(lngLine)
  End If
  
  'if using VB syntax,
  If agCurrentSyntax = lsVB Then
    'check for continuation
    If Right$(strCurrentLine, 2) = " _" And Len(strCurrentLine) >= 2 Then
      'set flag so nextchar treats the next line as continuation of current line
      blnContinueLine = True
      'strip off continuation characters
      strCurrentLine = Trim$(Left$(strCurrentLine, Len(strCurrentLine) - 2))
    Else
      blnContinueLine = False
    End If
  End If
End Sub
Private Function NextChar(Optional ByVal blnNoNewLine As Boolean = False) As String
  'gets the next non-space character (tabs (ascii code H&9, are converted
  'to a space character, and ignored) from the input stream if using c
  'syntax, or from the current line if using VB syntax
  
  'if the NoNewLine flag is passed,
  'the function will not look past current line for next
  'character; if no character on current line,
  'lngPos is set to end of current line, and
  'empty string is returned
  On Error GoTo ErrHandler
  
  'if already at end of input (lngLine=-1)
  If lngLine = -1 Then
    'just exit
    NextChar = vbNullString
    Exit Function
  End If
  
  Do
    'first, increment position
    lngPos = lngPos + 1
    'if past end of this line,
    If lngPos > Len(strCurrentLine) Then
      'if using VB syntax
      If agCurrentSyntax = lsVB Then
        'prevent getting a new line if continuation is NOT in effect
        blnNoNewLine = Not blnContinueLine
      End If
    
      'if can't get another line,
      If blnNoNewLine Then
        'move pointer back
        lngPos = lngPos - 1
        'return empty string
        NextChar = vbNullString
        Exit Function
      End If
      
      'get the next line
      IncrementLine
      'if at end of input
      If lngLine = -1 Then
        'exit with no character
        Exit Function
      End If
      'increment pointer(so it points to first character of line)
      lngPos = lngPos + 1
    End If
    
    NextChar = Mid$(strCurrentLine, lngPos, 1)
    'treat tabs as a space
    If NextChar = vbTab Then
      NextChar = " "
    End If
  Loop Until NextChar <> " " And LenB(NextChar) <> 0
Exit Function

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 656, strErrSrc, Replace(LoadResString(656), ARG1, CStr(lngError) & ":" & strError)
End Function
Private Function NextCommand(Optional ByVal blnNoNewLine As Boolean = False) As String
  'this function will return the next command, which is comprised
  'of command elements, and separated by element separators
  'command elements include:
  '  characters a-z, A-Z, numbers 0-9, and:  #$%.@_
  '  NOTE: inside quotations, ALL characters, including spaces
  '  are considered command elements
  '
  'element separators include:
  '  space, !"&'()*+,-/:;<=>?[\]^`{|}~
  '
  'element separators other than space are normally returned
  'as a single character command; there are some exceptions
  'where element separators will include additional characters:
  '  !=, &&, *=, ++, +=, --, -=, /=, //, <=, <>, =<, ==, =>, >=, ><, ||
  '
  'when a command starts with a quote, the command returns
  'after a closing quote is found, regardless of characters
  'inbetween the quotes
  '
  'if end of input is reached it returns empty string
  
  Dim intCmdEnd As Integer
  Dim intChar As Integer
  Dim blnInQuotes As Boolean
  
  'in VB syntax, commands are never retrieved from a new line
  If agCurrentSyntax = lsVB Then
    blnNoNewLine = True
  End If
  
  'find next non-blank character
  NextCommand = NextChar(blnNoNewLine)
  'if at end of input,
  If lngLine = -1 Then
    'return empty string
    NextCommand = vbNullString
    Exit Function
  End If
  'if no character returned
  If LenB(NextCommand) = 0 Then
    Exit Function
  End If
  
  On Error GoTo ErrHandler

  'if command is a element separator:
  Select Case AscW(NextCommand)
  Case 39, 40, 41, 44, 58, 59, 63, 91, 92, 93, 94, 96, 123, 125, 126 '  '(),:;?[\]^`{}~
    'return this single character as a command
    Exit Function
  Case 61 '=
    'special case; "=", "=<" and "=>" returned as separate commands
    Select Case Mid$(strCurrentLine, lngPos + 1, 1)
    Case "<", ">"
      'increment pointer
      lngPos = lngPos + 1
      'return the two byte cmd (swap so we get ">=" and "<="
      ' instead of "=>" and "=<"
      NextCommand = Mid$(strCurrentLine, lngPos, 1) & NextCommand
    
    Case "=" '"==" only returned if using c++ syntax
      If agCurrentSyntax = lsAGI Then
        'increment pointer
        lngPos = lngPos + 1
        'return the two byte cmd
        NextCommand = "=="
      End If
    End Select
    Exit Function
  Case 34 '"
    'special case; quote means start of a string
    blnInQuotes = True
  Case 43 '+
    'special case; "+", "++" and "+=" returned as separate commands
    If agCurrentSyntax = lsAGI Then
      If Mid$(strCurrentLine, lngPos + 1, 1) = "+" Then
        'increment pointer
        lngPos = lngPos + 1
        'return shorthand increment
        NextCommand = "++"
      ElseIf Mid$(strCurrentLine, lngPos + 1, 1) = "=" Then
        lngPos = lngPos + 1
        'return shorthand addition
        NextCommand = "+="
      End If
    End If
    Exit Function
  Case 45 '-
    'special case; "-", "--" and "-=" returned as separate commands
    If agCurrentSyntax = lsAGI Then
      If Mid$(strCurrentLine, lngPos + 1, 1) = "-" Then
        'increment pointer
        lngPos = lngPos + 1
        'return shorthand decrement
        NextCommand = "--"
      ElseIf Mid$(strCurrentLine, lngPos + 1, 1) = "=" Then
        lngPos = lngPos + 1
        'return shorthand subtract
        NextCommand = "-="
      End If
    End If
    Exit Function
  Case 33 '!
    'special case; "!" and "!=" returned as separate commands
    If agCurrentSyntax = lsAGI Then
      If Mid$(strCurrentLine, lngPos + 1, 1) = "=" Then
        'increment pointer
        lngPos = lngPos + 1
        'return not equal
        NextCommand = "!="
      End If
    End If
    Exit Function
  Case 60 '<
    'special case; "<", "<=" and "<>" returned as separate commands
    If Mid$(strCurrentLine, lngPos + 1, 1) = "=" Then
      'increment pointer
      lngPos = lngPos + 1
      'return less than or equal
      NextCommand = "<="
    ElseIf Mid$(strCurrentLine, lngPos + 1, 1) = ">" Then
      'increment pointer
      lngPos = lngPos + 1
      'return not equal
      NextCommand = "<>"
    End If
    Exit Function
  Case 62 '>
    'special case; ">", ">=" and "><" returned as separate commands
    If Mid$(strCurrentLine, lngPos + 1, 1) = "=" Then
      'increment pointer
      lngPos = lngPos + 1
      'return greater than or equal
      NextCommand = ">="
    ElseIf Mid$(strCurrentLine, lngPos + 1, 1) = "<" Then
      'increment pointer
      lngPos = lngPos + 1
      'return not equal ('><' is same as '<>')
      NextCommand = "<>"
    End If
    Exit Function
  Case 42 '*
    'special case; "*" and "*=" returned as separate commands;
    If agCurrentSyntax = lsAGI Then
      If Mid$(strCurrentLine, lngPos + 1, 1) = "=" Then
        'increment pointer
        lngPos = lngPos + 1
        'return shorthand multiplication
        NextCommand = "*="
        Exit Function
      End If
    End If
    Exit Function
  Case 47 '/
    'special case; "/" , "//" and "/=" returned as separate commands
    If agCurrentSyntax = lsAGI Then
      If Mid$(strCurrentLine, lngPos + 1, 1) = "=" Then
        lngPos = lngPos + 1
        'return shorthand division
        NextCommand = "/="
      ElseIf Mid$(strCurrentLine, lngPos + 1, 1) = "/" Then
        lngPos = lngPos + 1
        NextCommand = "//"
      End If
    End If
    Exit Function
  Case 124 '|
    'special case; "|" and "||" returned as separate commands
    If agCurrentSyntax = lsAGI Then
      If Mid$(strCurrentLine, lngPos + 1, 1) = "|" Then
        'increment pointer
        lngPos = lngPos + 1
        'return double '|'
        NextCommand = "||"
      End If
    End If
    Exit Function
  Case 38 '&
    'special case; "&" and "&&" returned as separate commands
    If agCurrentSyntax = lsAGI Then
      If Mid$(strCurrentLine, lngPos + 1, 1) = "&" Then
        'increment pointer
        lngPos = lngPos + 1
        'return double '&'
        NextCommand = "&&"
      End If
    End If
    Exit Function
  End Select
  
  'if not a text string,
  If Not blnInQuotes Then
    'continue adding characters until element separator or EOL is reached
    Do Until lngPos + 1 > Len(strCurrentLine)
'      'if at end of line
'      If lngPos + 1 > Len(strCurrentLine) Then
'        'end found;
'        Exit Do
'      End If
      intChar = AscW(Mid$(strCurrentLine, lngPos + 1, 1))
      Select Case intChar
      Case 32, 33, 34, 38 To 45, 47, 58 To 63, 91 To 94, 96, 123 To 126
        '  space, !"&'()*+,-/:;<=>?[\]^`{|}~
        'end of command text found;
        Exit Do
      Case Else
        'add character
        NextCommand = NextCommand & ChrW$(intChar)
        'incrmeent position
        lngPos = lngPos + 1
      End Select
    Loop 'While True
    
    'check for 'end if'
    If NextCommand = "end" Then
      If Left$(Trim$(Right$(strCurrentLine, Len(strCurrentLine) - lngPos)), 2) = "if" Then
        NextCommand = "end if"
        lngPos = InStr(lngPos, strCurrentLine, "i", vbTextCompare) + 1
      End If
    End If
  Else
    'if past end of line
    '(which could only happen if a line contains a single double quote on it)
    If lngPos + 1 > Len(strCurrentLine) Then
      'return the single quote
      Exit Function
    End If
      
    'add characters until another true quote is found
    Do
      'reset pointer to next
      intChar = AscW(Mid$(strCurrentLine, lngPos + 1, 1))
      'increment position
      lngPos = lngPos + 1
      
      'if the next character is a quotation mark,
      If intChar = 34 Then
        'if NOT an embedded quote (\")
        If Right$(NextCommand, 1) <> "\" Then
          'this quote marks end of string
          blnInQuotes = False
        End If
      End If
      
      NextCommand = NextCommand & ChrW$(intChar)
      
      'if at end of line
      If (lngPos = Len(strCurrentLine)) Then
        'if still in quotes,
        If blnInQuotes Then
          'set inquotes to false to exit the loop
          'the compiler will have to recognize that
          'this text string is not properly enclosed in quotes
          blnInQuotes = False
        End If
      End If
    Loop While blnInQuotes
  End If
Exit Function

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 657, strErrSrc, Replace(LoadResString(657), ARG1, CStr(lngError) & ":" & strError)
End Function
Private Function CompileIf() As Boolean
  'this routine will read and validate a group of test commands
  'for an 'if' statement and return
  'it is entered when the compiler encounters an 'if' command
  'the syntax expected for test commands is:
  '
  '   if(<test>){
  'or
  '   if(<test> &&/And <test>){
  'or
  '   if((<test> ||/Or <test>)){
  '
  '   <test> may be a test command (<tstcmd>(arg1, arg2, ..., argn)
  '   or a special syntax representation of a test command:
  '     fn        ==> isset(fn)
  '     vn == m   ==> equaln(vn,m)
  '     etc
  '
  'valid special comparison expressions are ==, <>, >, =<, <, >=
  'OR'ed tests must always be enclosed in parenthesis; AND'ed tests
  'must never be enclosed in parentheses (this ensures the compiled code
  'will be compatible with the AGI interpreter)
  '
  '(any test command may have the negation operator (!/Not) placed directly in front of it
  
  Dim strTestCmd As String, strArg As String
  Dim bytTestCmd As Byte, bytArg(7) As Byte
  Dim lngArg As Long, lngWord() As Long
  Dim intWordCount As Integer
  Dim i As Integer
  Dim blnIfBlock As Boolean
  Dim blnNeedNextCmd As Boolean
  Dim intNumTestCmds As Integer
  Dim intNumCmdsInBlock As Integer
  Dim blnNOT As Boolean
  
  On Error GoTo ErrHandler
  
  blnIfBlock = False
  intNumTestCmds = 0
  intNumCmdsInBlock = 0
  blnNeedNextCmd = True

  
  'write out starting if byte
  tmpLogRes.WriteByte &HFF
  
  'if using c++ syntax,
  If agCurrentSyntax = lsAGI Then
    'next character should be "("
    If NextChar() <> "(" Then
      blnError = True
      strErrMsg = LoadResString(4002)
      Exit Function
    End If
  End If
  
  'now, step through input, until final ')'' is found:
  Do
    'get next command
    strTestCmd = NextCommand()
    'check for end of input,
    If lngLine = -1 Then
      blnError = True
      strErrMsg = LoadResString(4106)
      Exit Function
    End If
    
    'if awaiting a test command,
    If blnNeedNextCmd Then
      Select Case strTestCmd
      Case "(" 'open paran
        'if already in a block
        If blnIfBlock Then
          blnError = True
          strErrMsg = LoadResString(4045)
          Exit Function
        End If
        'write 'or' block start
        tmpLogRes.WriteByte &HFC
        blnIfBlock = True
        intNumCmdsInBlock = 0
      Case ")"
        'if a test command is expected, ')' always causes error
        If intNumTestCmds = 0 Then
          strErrMsg = LoadResString(4057)
        ElseIf blnIfBlock And intNumCmdsInBlock = 0 Then
          strErrMsg = LoadResString(4044)
        Else
          strErrMsg = LoadResString(4056)
        End If
        blnError = True
        Exit Function
      Case Else
        'check for NOT
        blnNOT = (strTestCmd = strNot)
        If blnNOT Then
          tmpLogRes.WriteByte &HFD
          'read in next test command
          strTestCmd = NextCommand()
          'check for end of input,
          If lngLine = -1 Then
            blnError = True
            strErrMsg = LoadResString(4106)
            Exit Function
          End If
        End If
        bytTestCmd = CommandNum(True, strTestCmd)
        'if command not found,
        If bytTestCmd = 255 Then
          'check for special syntax
          If Not CompileSpecialIf(strTestCmd, blnNOT) Then
            'error; the CompileSpecialIf function
            'sets the error codes, and CompileLogic will
            'call the error handler
            Exit Function
          End If
        Else
          'write the test command code
          tmpLogRes.WriteByte bytTestCmd
          'next command should be "("
          If NextChar() <> "(" Then
            blnError = True
            strErrMsg = LoadResString(4048)
            Exit Function
          End If
          
          'if said command
          If bytTestCmd = &HE Then
            'enable error trapping to catch invalid word
            On Error Resume Next
            
            intWordCount = 0
            ReDim lngWord(0)
            'get first word arg
            lngArg = GetNextArg(atVocWrd, intWordCount + 1)
            'if error
            If blnError Then
              'exit
              Exit Function
            End If
            
            'loop to add this word, and any more
            Do
              'add this word number to array of word numbers
              ReDim Preserve lngWord(intWordCount)
              lngWord(intWordCount) = lngArg
              intWordCount = intWordCount + 1
              'if too many words
              If intWordCount = 10 Then
                blnError = True
                strErrMsg = LoadResString(4093)
                Exit Function
              End If
              
              'get next character
              '(should be a comma, or close parenthesis, if no more words)
              strArg = NextChar(True)
              If LenB(strArg) <> 0 Then
                Select Case AscW(strArg)
                Case 41 ')'
                  'move pointer back one space so
                  'the ')' will be found at end of command
                  lngPos = lngPos - 1
                  Exit Do
                  
                Case 44 ','
                  'expected; now check for next word argument
                  lngArg = GetNextArg(atVocWrd, intWordCount + 1)
                  'if error
                  If blnError Then
                    'exit
                    Exit Function
                  End If
                  
                Case Else
                  'error
                  blnError = True
                  strErrMsg = Replace(LoadResString(4047), ARG1, CStr(intWordCount))
                  Exit Function
                End Select
              Else
                'error
                blnError = True
                strErrMsg = Replace(LoadResString(4047), ARG1, CStr(intWordCount))
                Exit Function
              End If
            Loop While True
              
            'reset error handling
            On Error GoTo 0
            
            'need to write number of arguments for 'said'
            'before writing arguments themselves
            tmpLogRes.WriteByte CByte(intWordCount)
            
            'now add words
            For i = 0 To intWordCount - 1
              'write word value
              tmpLogRes.WriteWord lngWord(i)
            Next i
          Else
            'not 'said'; extract arguments for this command
            For i = 0 To agTestCmds(CByte(bytTestCmd)).ArgCount - 1
              'after first argument, verify comma separates arguments
              If i > 0 Then
                If NextChar(True) <> "," Then
                  blnError = True
                  strErrMsg = Replace(LoadResString(4047), ARG1, CStr(i))
                  Exit Function
                End If
              End If
              bytArg(i) = GetNextArg(agTestCmds(CByte(bytTestCmd)).ArgType(i), i)
              'if error
              If blnError Then
                Exit Function
              End If
              'write argument
              tmpLogRes.WriteByte bytArg(i)
            Next i
          End If
          'next character should be ")"
          If NextChar() <> ")" Then
            blnError = True
            strErrMsg = LoadResString(4048)
            Exit Function
          End If
          'validate arguments for this command
          If Not ValidateIfArgs(bytTestCmd, bytArg()) Then
            'error assigned by called function
            Exit Function
          End If
        End If
        
        'command added
        intNumTestCmds = intNumTestCmds + 1
        'if in IF block,
        If blnIfBlock Then
          intNumCmdsInBlock = intNumCmdsInBlock + 1
        End If
        'toggle off need for test command
        blnNeedNextCmd = False
      End Select
    Else 'not awaiting a test command
      Select Case strTestCmd
      Case "then" 'applicable to VB syntax only; in AGI syntax
                  'the closing parenthesis causes loop to exit
        If agCurrentSyntax <> lsVB Then
          'error; reserved word for vb; dont use in agi syntax either
          blnError = True
          strErrMsg = LoadResString(4107)
          Exit Function
        End If
        'if in an or block
        If blnIfBlock Then
          'error
          blnError = True
          strErrMsg = LoadResString(4004)
          Exit Function
        Else
          'ensure at least one command
          If intNumTestCmds = 0 Then
            blnError = True
            strErrMsg = LoadResString(4057)
            Exit Function
          End If
          'end of if statement found
          Exit Do
        End If
      Case strNot
        'invalid
        blnError = True
        strErrMsg = LoadResString(4097)
        Exit Function
      Case strAnd
        'if inside brackets
        If blnIfBlock Then
          blnError = True
          strErrMsg = LoadResString(4037)
          Exit Function
        End If
        blnNeedNextCmd = True
      Case strOr
        'if NOT inside brackets
        If Not blnIfBlock Then
          blnError = True
          strErrMsg = LoadResString(4100)
          Exit Function
        End If
        blnNeedNextCmd = True
      Case ")"
        'if inside brackets
        If blnIfBlock Then
          'ensure at least one command in block,
          If intNumCmdsInBlock = 0 Then
            blnError = True
            strErrMsg = LoadResString(4044)
            Exit Function
          End If
          'close brackets
          blnIfBlock = False
          tmpLogRes.WriteByte &HFC
        Else
          Select Case agCurrentSyntax
          Case lsVB
            'close parenthesis w/o open parenthesis causes error
            'error
            blnError = True
            strErrMsg = LoadResString(4058)
            Exit Function
          Case lsAGI
            'ensure at least one command in block,
            If intNumTestCmds = 0 Then
              blnError = True
              strErrMsg = LoadResString(4044)
              Exit Function
            End If
            'end of if found
            Exit Do
          End Select
        End If
      Case Else
        If blnIfBlock Then
          strErrMsg = LoadResString(4101)
        Else
          strErrMsg = LoadResString(4038)
        End If
        blnError = True
        Exit Function
      End Select
    End If
  'never leave loop normally; error, end of input, or successful
  'compilation of test commands will all exit loop directly
  Loop While True
  
  'write ending if byte
  tmpLogRes.WriteByte &HFF
  'return true
  CompileIf = True
Exit Function

ErrHandler:
  blnError = True
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "CompileIf")
  'err.clear
End Function
Private Function ConcatArg(strText As String) As String
  'this function concatenates strings; it assumes strText
  'is the string that was just read into the compiler;
  'it then checks if there are additional elements of
  'this string to add to it; if so, they are added.
  'it returns the completed string, with lngPos and lngLine
  'set accordingly (if there is nothing to concatenate, it
  'returns original string)
  'NOTE: input string is checked for starting and ending
  'quotation marks
  
  Dim strTextContinue As String
  Dim lngLastPos As Long, lngLastLine As Long
  Dim strLastLine As String
  
  On Error GoTo ErrHandler
  
  'verify at least two characters
  If Len(strText) < 2 Then
    'error
    blnError = True
    strErrMsg = LoadResString(4081)
    Exit Function
  End If
  
  'start with input string
  ConcatArg = strText
  
  'check msg for quotes
  If Left$(ConcatArg, 1) <> QUOTECHAR Or Right$(ConcatArg, 1) <> QUOTECHAR Then
    Select Case agMainLogSettings.ErrorLevel
    Case leHigh
      blnError = True
      strErrMsg = LoadResString(4051)
      Exit Function
    Case leMedium
      'add quotes as appropriate
      If AscW(ConcatArg) <> 34 Then
        ConcatArg = QUOTECHAR & ConcatArg
      End If
      If AscW(Right$(ConcatArg, 1)) <> 34 Then
        ConcatArg = ConcatArg & QUOTECHAR
      End If
      'set warning
      AddWarning LoadResString(5002)
    Case leLow
      'add quotes as appropriate
      If AscW(ConcatArg) <> 34 Then
        ConcatArg = QUOTECHAR & ConcatArg
      End If
      If AscW(Right$(ConcatArg, 1)) <> 34 Then
        ConcatArg = ConcatArg & QUOTECHAR
      End If
    End Select
  End If
  
  'save current position info
  lngLastPos = lngPos
  lngLastLine = lngLine
  strLastLine = strCurrentLine
  'get next command
  strTextContinue = NextCommand()
  
  Select Case agCurrentSyntax
  Case lsVB
    'add strings until concatenation is complete
    Do Until strTextContinue <> "&"
      'get string to concatenate
      strTextContinue = NextCommand()
      'if it is not a string
      If Left$(strTextContinue, 1) <> QUOTECHAR Or Right$(strTextContinue, 1) <> QUOTECHAR Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          'error
          blnError = True
          strErrMsg = LoadResString(4080)
          Exit Function
        Case leMedium
          'add quotes as appropriate
          If LenB(strTextContinue) <> 0 Then
            If AscW(strTextContinue) <> 34 Then
              strTextContinue = QUOTECHAR & strTextContinue
            End If
            If AscW(Right$(strTextContinue, 1)) <> 34 Then
              strTextContinue = strTextContinue & QUOTECHAR
            End If
          Else
            'error - nothing to concatenate
            'this could only happen if source code ended with
            'a concatenation that was improperly set up
            blnError = True
            strErrMsg = LoadResString(4080)
            Exit Function
          End If
          'set warning
          AddWarning LoadResString(5003)
        Case leLow
          'add quotes as appropriate
          If AscW(strTextContinue) <> 34 Then
            strTextContinue = QUOTECHAR & strTextContinue
          End If
          If AscW(Right$(strTextContinue, 1)) <> 34 Then
            strTextContinue = strTextContinue & QUOTECHAR
          End If
        End Select
      End If
      'strip off ending quote of current msg
      ConcatArg = Left$(ConcatArg, Len(ConcatArg) - 1)
      'add it to strText
      ConcatArg = ConcatArg & Right$(strTextContinue, Len(strTextContinue) - 1)
      'save current position info
      lngLastPos = lngPos
      lngLastLine = lngLine
      strLastLine = strCurrentLine
      'get next command
      strTextContinue = NextCommand()
    Loop
  Case lsAGI
    'if at end of last line
    If lngLastPos = Len(strLastLine) Then
      'add strings until concatenation is complete
      Do Until Left$(strTextContinue, 1) <> QUOTECHAR
        'if not a closing quote
        If Right$(strTextContinue, 1) <> QUOTECHAR Then
          Select Case agMainLogSettings.ErrorLevel
          Case leHigh
            'error
            blnError = True
            strErrMsg = LoadResString(4080)
            Exit Function
          Case leMedium
            'add quote
            strTextContinue = strTextContinue & QUOTECHAR
            'set warning
            AddWarning LoadResString(5002)
          Case leLow
            'add quote
            strTextContinue = strTextContinue & QUOTECHAR
          End Select
        End If
        'strip off ending quote of current msg
        ConcatArg = Left$(ConcatArg, Len(ConcatArg) - 1)
        'add it to strText
        ConcatArg = ConcatArg & Right$(strTextContinue, Len(strTextContinue) - 1)
        'save current position info
        lngLastPos = lngPos
        lngLastLine = lngLine
        strLastLine = strCurrentLine
        'get next command
        strTextContinue = NextCommand()
      Loop
    End If
  End Select
  
  'after end of string found, move back to correct position
  lngPos = lngLastPos
  lngLine = lngLastLine
  strCurrentLine = strLastLine
Exit Function

ErrHandler:
  'raise an error
  blnError = True
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "ConcatArg")
End Function


Private Function RemoveComments() As Boolean
  'this function strips out comments from the input text
  'and trims off leading and trailing spaces
  '
  'agi comments:
  '      // - rest of line is ignored
  '      [ - rest of line is ignored
  '      /* Text between these are ignored */
  '
  '      The /*...*/ can be nested
  '
  'VB comments:
  '      ' - rest of line is ignored
  '      Rem - rest of line is ignored
  

  Dim intChar As Integer
  Dim blnInQuotes As Boolean, blnLastLineContinued As Boolean
  Dim intCommentDepth As Integer
  Dim intCommentStart As Integer
  Dim intROLIgnore As Integer
  
  On Error GoTo ErrHandler

  'initialize variables
  intCommentDepth = 0
  
  'reset compiler
  ResetCompiler
  lngIncludeOffset = 0
  lngLine = 0
  
  Do
    'if continuing from last line and commenting out
    'AND in agi syntax (VB NEVER continues comments over multiple lines)
    If blnLastLineContinued And intROLIgnore > 0 And agCurrentSyntax = lsAGI Then
      'ignore all of this line as well
      intROLIgnore = 1
    Else
      'reset rol ignore
      intROLIgnore = 0
    End If
    
    'reset comment start & char ptr, and inquotes
    intCommentStart = 1
    intChar = 0
    blnInQuotes = False
    
    'if this line is not empty,
    If LenB(strCurrentLine) <> 0 Then
      Do Until intChar >= Len(strCurrentLine)
        'get next character from string
        intChar = intChar + 1
        'if NOT inside a quotation,
        If Not blnInQuotes Then
          Select Case agCurrentSyntax
          Case lsVB
            'search for comment character OR whole word (REM)
            If (((Mid$(strCurrentLine, intChar, Len(strCommentChar2)) = strCommentChar2) And (FindWholeWord(intChar, strCurrentLine, strCommentChar2))) Or (Mid$(strCurrentLine, intChar, 1) = strCommentChar1)) And (intROLIgnore = 0) And intCommentDepth = 0 Then
              intROLIgnore = intChar
              Exit Do
            End If
          Case lsAGI
            'search for comment characters
            If ((Mid$(strCurrentLine, intChar, Len(strCommentChar2)) = strCommentChar2) Or (Mid$(strCurrentLine, intChar, 1) = strCommentChar1)) And (intROLIgnore = 0) And intCommentDepth = 0 Then
              intROLIgnore = intChar
            End If
            'check for comment start
            If (Mid$(strCurrentLine, intChar, 2) = strCommentStart) And intROLIgnore = 0 Then
              'if this first comment block
              If intCommentDepth = 0 Then
                intCommentStart = intChar
              End If
              intCommentDepth = intCommentDepth + 1
            'check for end of comment
            ElseIf (Mid$(strCurrentLine, intChar, 2) = strCommentEnd) Then
              'if this end block comes BEFORE a start comment block,
              If intCommentDepth = 0 Then
                'set depth and location, so it will be treated correctly as an end of comment block
                'net effect will be to ignore the "*/" characters only
                intCommentDepth = 1
                intCommentStart = intChar
              End If
              intCommentDepth = intCommentDepth - 1
              'if made back to end of comment depth,
              If intCommentDepth = 0 Then
                'remove the comment from the line
                strCurrentLine = Left$(strCurrentLine, intCommentStart - 1) & Right$(strCurrentLine, Len(strCurrentLine) - (intChar + 1))
                'reset current character to next character after comment is removed
                intChar = intCommentStart - 1
                'if there was a previous comment on this line requiring ignoring the rest of line,
                If intROLIgnore > 0 Then
                  'reset the ignore from position
                  intROLIgnore = intROLIgnore - (intChar - intCommentStart + 2)
                End If
              End If
            End If
          End Select
          'set quote flag
          If intChar = 0 Then
            blnInQuotes = False
          Else
            blnInQuotes = (AscW(Mid$(strCurrentLine, intChar)) = 34)
          End If
        Else
          'check to see if a quotation is starting or ending
          If AscW(Mid$(strCurrentLine, intChar)) = 34 Then
            Debug.Assert intChar > 1
            'if there is NOT a backslash before the quote
            If AscW(Mid$(strCurrentLine, intChar - 1)) <> 92 Then
              'not a real quote; toggle back
              blnInQuotes = Not blnInQuotes
            End If
          End If
        End If
      Loop
      'if in the middle of a comment block,
      If intCommentDepth > 0 Then
        strCurrentLine = Left$(strCurrentLine, intCommentStart - 1)
      End If
      'if any part of line should be ignored,
      If intROLIgnore > 0 Then
        strCurrentLine = Left$(strCurrentLine, intROLIgnore - 1)
      End If
    End If
    'replace the line with the stripped line
    ReplaceLine Trim$(strCurrentLine), intROLIgnore > 0
    
    'if line will be continued on next line
    blnLastLineContinued = blnContinueLine
    
    'get next line
    IncrementLine
  Loop Until lngLine = -1
  
  'if an open comment block was found
  If intCommentDepth > 0 Then
    Select Case agMainLogSettings.ErrorLevel
    Case leHigh
      strErrMsg = LoadResString(4046)
      Exit Function
    Case leMedium
      'assume end of block occurred at end of input
      AddWarning LoadResString(5012)
    Case leLow
      'assume end of block occurred at end of input
    End Select
  End If
    
  'success
  RemoveComments = True
Exit Function
ErrHandler:
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "RemoveComments")
  Err.Clear
End Function

Private Function AddIncludes(stlLogicText As StringList) As Boolean
  'this function uses the logic text that is passed to the compiler
  'to create the input text that is parsed.
  'it copies the lines from the logic text to the input text, and
  'replaces any include file lines with the actual lines from the
  'include file (include file lines are given a 'header' to identify
  'them as include lines)
  
  
  Dim IncludeLines As StringList
  Dim strIncludeFilename As String
  Dim strIncludePath As String
  Dim strIncludeText As String
  Dim CurIncludeLine As Long   ' current line in IncludeLines (the include file)
  Dim intFileCount As Integer
  Dim i As Integer
  Dim intFile As Integer
  Dim lngLineCount As Long
  
  On Error GoTo ErrHandler
  
  Set stlInput = New StringList
  Set IncludeLines = New StringList 'only temporary,
  
  lngLine = 0
  lngLineCount = stlLogicText.Count
  
  'module is always main module
  strModule = vbNullString
  
  'step through all lines
  Do
    'set errline
    lngErrLine = lngLine
    'check this line for include statement
    If Left$(stlLogicText(lngLine), 8) = "#include" Then
      'proper format requires a space after 'include'
      If Mid$(stlLogicText(lngLine), 9, 1) <> " " Then
        'generate error
        strErrMsg = LoadResString(4103)
        Exit Function
      End If
      'build include filename
      strIncludeFilename = Trim$(Right$(stlLogicText(lngLine), Len(stlLogicText(lngLine)) - 9))
      
      'check for a filename
      If LenB(strIncludeFilename) = 0 Then
        strErrMsg = LoadResString(4060)
        Exit Function
      End If
      
      'if quotes aren't used correctly
      If Left$(strIncludeFilename, 1) = QUOTECHAR And Right$(strIncludeFilename, 1) <> QUOTECHAR Or _
         Right$(strIncludeFilename, 1) = QUOTECHAR And Left$(strIncludeFilename, 1) <> QUOTECHAR Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          'return error: improper use of quote marks
          strErrMsg = LoadResString(4059)
          Exit Function
        Case leMedium, leLow
          'assume quotes are needed
          If AscW(strIncludeFilename) <> 34 Then
            strIncludeFilename = QUOTECHAR & strIncludeFilename
          End If
          If AscW(Right$(strIncludeFilename, 1)) <> 34 Then
            strIncludeFilename = strIncludeFilename & QUOTECHAR
          End If
          'set warning
          AddWarning "(5028) improper use of quote marks around Include file. Include filename assumed to be: " & strIncludeFilename
        End Select
      End If
      
      'if quotes,
      If Left$(strIncludeFilename, 1) = QUOTECHAR Then
        'strip off quotes
        strIncludeFilename = Mid$(strIncludeFilename, 2, Len(strIncludeFilename) - 2)
      End If
      
      'if filename doesnt include a path,
      If LenB(JustPath(strIncludeFilename, True)) = 0 Then
        'get full path name to include file
        strIncludeFilename = agResDir & strIncludeFilename
      End If
      
      'verify file exists
      If Not FileExists(strIncludeFilename) Then
        strErrMsg = Replace(LoadResString(4050), ARG1, strIncludeFilename)
        Exit Function
      End If
'****
'      cant check for open includes; they are in a different application
'****

      On Error Resume Next
      'now open the include file, and get the text
      intFile = FreeFile()
      Open strIncludeFilename For Binary As intFile
      strIncludeText = String$(LOF(intFile), 0)
      Get intFile, 1, strIncludeText
      Close intFile
      'check for error,
      If Err.Number <> 0 Then
        strErrMsg = Replace(LoadResString(4055), ARG1, strIncludeFilename)
        Exit Function
      End If
      
      On Error GoTo ErrHandler
      
      'assign text to stringlist
      Set IncludeLines = New StringList
      IncludeLines.Assign strIncludeText
      
      'if there are any lines,
      If IncludeLines.Count > 0 Then
        'save file name to allow for error checking
        intFileCount = intFileCount + 1
        ReDim Preserve strIncludeFile(intFileCount)
        strIncludeFile(intFileCount) = strIncludeFilename
        
        'add all these lines into this position
        For CurIncludeLine = 0 To IncludeLines.Count - 1
          'verify the include file contains no includes
          If Left$(Trim$(IncludeLines(CurIncludeLine)), 2) = "#i" Then
            strErrMsg = LoadResString(4061)
            lngErrLine = CurIncludeLine
            Exit Function
          End If
          'include filenumber and line number from includefile
          stlInput.Add "#I" & CStr(intFileCount) & ":" & CStr(CurIncludeLine) & "#" & IncludeLines(CurIncludeLine)
        Next CurIncludeLine
      End If
      'add a blank line as a place holder for the 'include' line
      '(to keep line counts accurate when calculating line number for errors)
      stlInput.Add vbNullString
    Else
      'not an include line
      'check for any instances of #I, since these will
      'interfere with include line handling
      If Left$(stlLogicText(lngLine), 2) = "#i" Then
        strErrMsg = LoadResString(4069)
        Exit Function
      End If
      'copy the line by itself
      stlInput.Add stlLogicText(lngLine)
    End If
    lngLine = lngLine + 1
  Loop Until lngLine >= lngLineCount
  'done
  Set IncludeLines = Nothing
  'return success
  AddIncludes = True
Exit Function

ErrHandler:
  'unknown error
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "AddIncludes")
  Err.Clear
End Function

  

Private Function ReadDefines() As Boolean
  
  Dim i As Long, j As Integer
  Dim blnInQuote As Boolean
  Dim tdNewDefine As TDefine
  Dim rtn As Long
  
  On Error GoTo ErrHandler
  
  'reset count of defines
  lngDefineCount = 0
  
  'reset compiler
  ResetCompiler
  
  'reset error string
  strErrMsg = vbNullString
  
  'step through all lines and find define values
  Do
    'check for define statement
    If InStr(1, strCurrentLine, strConst) = 1 Then
      'strip off define keyword
      strCurrentLine = Trim$(Right$(strCurrentLine, Len(strCurrentLine) - Len(strConst)))
      
      'if equal marker not present
      If InStrB(1, strCurrentLine, strConstEqual) = 0 Then
        'error
        strErrMsg = LoadResString(4104)
        Exit Function
      End If
      
      'split it by position of first space
      tdNewDefine.Name = Trim$(Left$(strCurrentLine, InStr(1, strCurrentLine, strConstEqual) - 1))
      tdNewDefine.Value = Trim$(Right$(strCurrentLine, Len(strCurrentLine) - InStr(1, strCurrentLine, strConstEqual)))
            
      'validate define name
      rtn = ValidateDefName(tdNewDefine.Name)
      'name error 7-12  are only warnings if error level is medium or low
      Select Case agMainLogSettings.ErrorLevel
      Case leMedium
        'check for name warnings
        Select Case rtn
        Case 7
          'set warning
          AddWarning Replace(LoadResString(5034), ARG1, tdNewDefine.Name)
          'reset return error code
          rtn = 0
         
        Case 8 To 12
          'set warning
          AddWarning Replace(LoadResString(5035), ARG1, tdNewDefine.Name)
          'reset return error code
          rtn = 0
        End Select
      Case leLow
        'check for warnings
        If rtn >= 7 And rtn <= 12 Then
          'reset return error code
          rtn = 0
        End If
      End Select
      
      'check for errors
      If rtn <> 0 Then
        'check for name errors
        Select Case rtn
        Case 1 ' no name
          strErrMsg = LoadResString(4070)
        Case 2 ' name is numeric
          strErrMsg = LoadResString(4072)
        Case 3 ' name is command
          strErrMsg = Replace(LoadResString(4021), ARG1, tdNewDefine.Name)
        Case 4 ' name is test command
          strErrMsg = Replace(LoadResString(4022), ARG1, tdNewDefine.Name)
        Case 5 ' name is a compiler keyword
          strErrMsg = Replace(LoadResString(4013), ARG1, tdNewDefine.Name)
        Case 6 ' name is an argument marker
          strErrMsg = LoadResString(4071)
        Case 7 ' name is already globally defined
          strErrMsg = Replace(LoadResString(4019), ARG1, tdNewDefine.Name)
        Case 8 ' name is reserved variable name
          strErrMsg = Replace(LoadResString(4018), ARG1, tdNewDefine.Name)
        Case 9 ' name is reserved flag name
          strErrMsg = Replace(LoadResString(4014), ARG1, tdNewDefine.Name)
        Case 10 ' name is reserved number constant
          strErrMsg = Replace(LoadResString(4016), ARG1, tdNewDefine.Name)
        Case 11 ' name is reserved object constant
          strErrMsg = Replace(LoadResString(4017), ARG1, tdNewDefine.Name)
        Case 12 ' name is reserved message constant
          strErrMsg = Replace(LoadResString(4015), ARG1, tdNewDefine.Name)
        Case 13 ' name contains improper character
          strErrMsg = LoadResString(4067)
        End Select
        'don't exit; check for define value errors first
      End If
      
      'validate define value
      rtn = ValidateDefValue(tdNewDefine)
      'value errors 4,5,6 are only warnings if error level is medium or low
      Select Case agMainLogSettings.ErrorLevel
      Case leMedium
        'if value error is due to missing quotes
        Select Case rtn
        Case 4  'string value missing quotes
          'fix the define value
          If AscW(tdNewDefine.Value) <> 34 Then
            tdNewDefine.Value = QUOTECHAR & tdNewDefine.Value
          End If
          If AscW(Right$(tdNewDefine.Value, 1)) <> 34 Then
            tdNewDefine.Value = tdNewDefine.Value & QUOTECHAR
          End If
          
          'set warning
          AddWarning LoadResString(5022)
          'reset error code
          rtn = 0
        Case 5 ' value is already defined by a reserved name
          'set warning
          AddWarning Replace(LoadResString(5032), ARG1, tdNewDefine.Value)
          'reset error code
          rtn = 0
          
        Case 6 ' value is already defined by a global name
          'set warning
          AddWarning Replace(LoadResString(5031), ARG1, tdNewDefine.Value)
          'reset error code
          rtn = 0
        End Select
      Case leLow
        'if value error is due to missing quotes
        Select Case rtn
        Case 4
          'fix the define value
          If AscW(tdNewDefine.Value) <> 34 Then
            tdNewDefine.Value = QUOTECHAR & tdNewDefine.Value
          End If
          If AscW(Right$(tdNewDefine.Value, 1)) <> 34 Then
            tdNewDefine.Value = tdNewDefine.Value & QUOTECHAR
          End If
          'reset return value
          rtn = 0
        Case 5, 6
          'reset return value
          rtn = 0
        End Select
      End Select
      
      'check for errors
      If rtn <> 0 Then
        'if already have a name error
        If LenB(strErrMsg) <> 0 Then
          'append value error
          strErrMsg = strErrMsg & "; and "
        End If
      
        'check for value error
        Select Case rtn
        Case 1 ' no value
          strErrMsg = strErrMsg & LoadResString(4073)
'
'a return value of 2 is no longer possible; this
'value has been removed from the ValidateDefineValue function
'        Case 2 ' value is an invalid argument marker
'          strErrMsg = strErrMsg & "4065: Invalid argument declaration value"
        Case 3 ' value contains an invalid argument value
          strErrMsg = strErrMsg & LoadResString(4042)
        Case 4 ' value is not a string, number or argument marker
          strErrMsg = strErrMsg & LoadResString(4082)
        Case 5 ' value is already defined by a reserved name
          strErrMsg = strErrMsg & Replace(LoadResString(4041), ARG1, tdNewDefine.Value)
        Case 6 ' value is already defined by a global name
          strErrMsg = strErrMsg & Replace(LoadResString(4040), ARG1, tdNewDefine.Value)
        End Select
      End If
      
      'if an error was generated during define validation
      If LenB(strErrMsg) <> 0 Then
        Exit Function
      End If
      
      'check all previous defines
      For i = 0 To lngDefineCount - 1
        If tdNewDefine.Name = tdDefines(i).Name Then
          strErrMsg = Replace(LoadResString(4012), ARG1, tdDefines(i).Name)
          Exit Function
        End If
        If tdNewDefine.Value = tdDefines(i).Value Then
          'numeric duplicates aren't a problem
          If Not IsNumeric(tdNewDefine.Value) Then
            Select Case agMainLogSettings.ErrorLevel
            Case leHigh
              'set error
              strErrMsg = Replace(Replace(LoadResString(4023), ARG1, tdDefines(i).Value), ARG2, tdDefines(i).Name)
              Exit Function
            Case leMedium
              'set warning
              AddWarning Replace(Replace(LoadResString(5033), ARG1, tdNewDefine.Value), ARG2, tdDefines(i).Name)
            Case leLow
              'do nothing
            End Select
          End If
        End If
      Next i
      
      'check define against labels
      If bytLabelCount > 0 Then
        For i = 1 To bytLabelCount
          If tdNewDefine.Name = llLabel(i).Name Then
            strErrMsg = Replace(LoadResString(4020), ARG1, tdNewDefine.Name)
            Exit Function
          End If
        Next i
      End If
           
      'save this define
      ReDim Preserve tdDefines(lngDefineCount)
      tdDefines(lngDefineCount) = tdNewDefine

      'increment counter
      lngDefineCount = lngDefineCount + 1
      
      'now set this line to empty so Compiler doesn"t try to read it
      If Left$(stlInput(lngLine), 2) = "#I" Then
        'this is an include line; need to leave include line info
        stlInput(lngLine) = Left$(stlInput(lngLine), InStr(4, stlInput(lngLine), "#"))
      Else
        'just blank out entire line
        stlInput(lngLine) = vbNullString
      End If
    End If
    'get next line
    IncrementLine
  Loop Until lngLine = -1
  
  ReadDefines = True
Exit Function

ErrHandler:
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "ReadDefines")
  Err.Clear
  Resume Next
End Function
Private Function ReadMsgs() As Boolean

  'note that stripped message lines also strip out the include header string
  'this doesn't matter since they are only blank lines anyway
  'only need include header info if error occurs, and errors never occur on
  'blank line
  
  Dim intMsgNum As Integer
  Dim strCmd As String, strMsgContinue As String
  Dim strMsgSep As String
  Dim lngMsgStart As Long
  Dim intMsgLineCount As Integer
  
  On Error GoTo ErrHandler
  
  'build blank message list
  For intMsgNum = 0 To 255
    strMsg(intMsgNum) = vbNullString
    blnMsg(intMsgNum) = False
  Next intMsgNum
  
  'reset compiler
  ResetCompiler
  
  Do
    'get first command on this line
    strCmd = NextCommand(True)
    
    'if this is the message marker
    If strCmd = strMsgMarker Then
      'save starting line number (incase this msg uses multiple lines)
      lngMsgStart = lngLine
      
      'if using VB syntax
      If agCurrentSyntax = lsVB Then
        'next character should be '('
        If NextChar() <> "(" Then
          'error
          strErrMsg = LoadResString(4095)
          Exit Function
        End If
      End If
      
      'get next command on this line
      strCmd = NextCommand(True)
      
      'this should be a msg number
      If Not IsNumeric(strCmd) Then
        'error
        strErrMsg = LoadResString(4077)
        Exit Function
      End If
      'validate msg number
      intMsgNum = VariableValue(strCmd)
      If intMsgNum <= 0 Then
        'error
        strErrMsg = LoadResString(4077)
        Exit Function
      End If
      'if msg is already assigned
      If blnMsg(intMsgNum) Then
        strErrMsg = Replace(LoadResString(4094), ARG1, CStr(intMsgNum))
        Exit Function
      End If
      
      'if using VB syntax
      If agCurrentSyntax = lsVB Then
        'next character should be ')'
        If NextChar() <> ")" Then
          'error
          strErrMsg = LoadResString(4005)
          Exit Function
        End If
        'next command should be equal sign
        If NextCommand() <> strMsgEqual Then
          'error
          strErrMsg = LoadResString(4035)
          Exit Function
        End If
      End If
      
      'get next string command
      strCmd = NextCommand(False)
      'concatenate, if necessary
      strCmd = ConcatArg(strCmd)
      'if error,
      If blnError Then
        Exit Function
      End If
      
      'nothing allowed after msg declaration
      If lngPos <> Len(strCurrentLine) Then
        'error
        strErrMsg = LoadResString(4099)
        Exit Function
      End If
  
      'strip off quotes
      strCmd = Mid$(strCmd, 2, Len(strCmd) - 2)
      
      'add msg
      strMsg(intMsgNum) = strCmd
      blnMsg(intMsgNum) = True
      
      Do
        'set this line to empty so compiler doesn't try to read it
        stlInput(lngMsgStart) = vbNullString
        'increment the counter (to get multiple lines, if string is
        'concatenated over more than one line)
        lngMsgStart = lngMsgStart + 1
      'continue until back to current line
      Loop Until lngMsgStart > lngLine
        
    End If
    'get next line
    IncrementLine
  Loop Until lngLine = -1
  ReadMsgs = True
Exit Function

ErrHandler:
  blnError = True
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "ReadMsgs")
  Err.Clear
End Function


Private Sub ReplaceLine(strNewLine As String, Optional DontContinue As Boolean = False)
  'this function replaces the current line in the input string
  'with the strNewLine, while preserving include header info and
  'continuation characters
  
  Dim strInclude As String, strContinue As String
  
    'if this is from an include file
    If Left$(stlInput(lngLine), 2) = "#I" Then
      'need to save include header info so it can
      'be preserved after comments are removed
      strInclude = Left$(stlInput(lngLine), InStr(2, stlInput(lngLine), "#"))
    Else
      strInclude = vbNullString
    End If
    
    'if in VB syntax AND this line continues on to next AND not forcing skip of continue
    If agCurrentSyntax = lsVB And Right$(stlInput(lngLine), 2) = " _" And Not DontContinue Then
      strContinue = " _"
    Else
      strContinue = vbNullString
    End If
    
    'replace the line
    stlInput(lngLine) = strInclude & strNewLine & strContinue
End Sub


Private Sub ResetCompiler()
  'resets the compiler so it points to beginning of input
  'also loads first line into strCurrentLine
  
  'reset include offset, so error trapping
  'can correctly count lines
  lngIncludeOffset = 0
  
  'reset continue flag and error flag
  blnContinueLine = False
  blnError = False

  'set line pointer to -2 so first call to
  'IncrementLine gets first line
  lngLine = -2
  
  'get first line
  IncrementLine
  'NOTE: don't need to worry about first line;
  'compiler has already verified the input has at least one line
End Sub


Public Sub SetResourceIDs()
  'builds array of resourceIDs so
  'convertarg function can iterate through them much quicker
  
  Dim tmpLog As AGILogic, tmpPic As AGIPicture
  Dim tmpSnd As AGISound, tmpView As AGIView
  
  If blnSetIDs Then
    Exit Sub
  End If
  
  ReDim strLogID(255)
  ReDim strPicID(255)
  ReDim strSndID(255)
  ReDim strViewID(255)
  
  For Each tmpLog In agLogs
    strLogID(tmpLog.Number) = tmpLog.ID
  Next
  
  For Each tmpPic In agPics
    strPicID(tmpPic.Number) = tmpPic.ID
  Next
  
  For Each tmpSnd In agSnds
    strSndID(tmpSnd.Number) = tmpSnd.ID
  Next
  
  For Each tmpView In agViews
    strViewID(tmpView.Number) = tmpView.ID
  Next
  
  'set flag
  blnSetIDs = True
End Sub

Public Sub SetSyntax(ByVal LogicSyntax As LogicSyntaxType)
  
  'assign current syntax
  agCurrentSyntax = LogicSyntax
  
  'if predefines aren't correct
  If agPreDefSyntax <> agCurrentSyntax Then
    AssignReservedDefines agCurrentSyntax
  End If
  
  'ensure default globals set for this syntax
  
  Select Case LogicSyntax
  Case lsVB
    strCommentStart = vbNullString
    strCommentEnd = vbNullString
    strCommentChar1 = Chr$(39)
    strCommentChar2 = "rem"
    strMsgEqual = "="
    strMsgMarker = "message"
    strConst = "const "
    strConstEqual = "="
    strEqual = "="
    strNotEqual = "<>"
    strAnd = "and"
    strOr = "or"
    strNot = "not"
  Case lsAGI
    strCommentStart = "/*"
    strCommentEnd = "*/"
    strCommentChar1 = "["
    strCommentChar2 = "//"
    strMsgEqual = vbNullString
    strMsgMarker = "#message"
    strConst = "#define "
    strConstEqual = " "
    strEqual = "=="
    strNotEqual = "!="
    strAnd = "&&"
    strOr = "||"
    strNot = "!"
  End Select
  
  'if a game is loaded,
  If agMainGame.GameLoaded Then
    'insert current values for reserved defines
    'agResDef(0).Value = "Ego"  'this one doesn't change
    agResDef(1).Value = QUOTECHAR & agMainGame.GameVersion & QUOTECHAR
    agResDef(2).Value = QUOTECHAR & agMainGame.GameAbout & QUOTECHAR
    agResDef(3).Value = QUOTECHAR & agMainGame.GameID & QUOTECHAR
    If agInvObj.Loaded Then
      'count of ACTUAL useable objects is one less than inventory object count
      'because the first object ('?') is just a placeholder
      agResDef(4).Value = agInvObj.Count - 1
    Else
      agResDef(4).Value = -1
    End If
  Else
    'if a game is not loaded, set reserved defines to defaults
    'agResDef(0).Value = "Ego"  'this one doesn't change
    agResDef(1).Value = vbNullString
    agResDef(2).Value = vbNullString
    agResDef(3).Value = vbNullString
    agResDef(4).Value = -1
    'agResDef(5).Value = "s0"  'this one doesn't change
  End If
End Sub

Private Sub AddWarning(ByVal WarningMsg As String)
  
  'set warning
  If blnWarnings Then
    strWarning = strWarning & "|"
  End If
  blnWarnings = True
  
  strWarning = strWarning & "Line " & CStr(lngErrLine + 1) & _
               IIf(LenB(strModule) <> 0, " in " & JustFileName(strModule), vbNullString) & _
               " - " & WarningMsg
End Sub

Private Function ValidateArgs(ByVal CmdNum As Long, ByRef ArgVal() As Byte) As Boolean

  'check for specific command issues
  On Error GoTo ErrHandler
  
  'for commands that can affect variable values, need to check against reserved variables
  'for commands that can affect flags, need to check against reserved flags
  
  'for other commands, check the passed arguments to see if values are appropriate
  
  Select Case CmdNum
  Case 1 To 8, 10, 165 To 168 'increment, decrement, assignv, addn, addv, subn, subv
                              'rindirect, mul.n, mul.v, div.n, div.v
    CheckResVarUse ArgVal(0), 255
    
    'for div.n(vA, B) only, check for divide-by-zero
    If CmdNum = 167 Then
      If ArgVal(1) = 0 Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          strErrMsg = LoadResString(4149)
          Exit Function
        Case leMedium
          AddWarning LoadResString(5030)
        Case leLow
        End Select
      End If
    End If
  
  Case 3 'assignn
    CheckResVarUse ArgVal(0), ArgVal(1)
  
  Case 12, 13, 14 'set, reset, toggle
    'check for reserved flags
    CheckResFlagUse ArgVal(0)
  
  Case 18 'new.room(A)
    'validate that this logic exists
    If Not agLogs.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4120)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5053)
      Case leLow
      End Select
    End If
    
  Case 20 'load.logics(A)
    'validate that logic exists
    If Not agLogs.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4121), ARG1, CStr(ArgVal(0)))
        Exit Function
      Case leMedium
        AddWarning LoadResString(5013)
      Case leLow
      End Select
    End If
  
  Case 22  'call(A)
    If ArgVal(0) = 0 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4118)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5010)
      Case leLow
        'no action
      End Select
    End If
    
  Case 30 'load.view(A)
    'validate that view exists
    If Not agViews.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4122), ARG1, CStr(ArgVal(0)))
        Exit Function
      Case leMedium
        AddWarning LoadResString(5015)
      Case leLow
      End Select
    End If
  
  Case 32 'discard.view(A)
    'validate that view exists
    If Not agViews.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4123), ARG1, CStr(ArgVal(0)))
        Exit Function
      Case leMedium
        AddWarning LoadResString(5024)
      Case leLow
      End Select
    End If
  
  Case 37 'position(oA, X,Y)
    'check x/y against limits
    If ArgVal(1) > 159 Or ArgVal(2) > 167 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
       strErrMsg = LoadResString(4128)
       Exit Function
      Case leMedium
        AddWarning LoadResString(5023)
      Case leLow
      End Select
    End If
    
  Case 39 'get.posn
    'check both values
    CheckResVarUse ArgVal(1), 255
    CheckResVarUse ArgVal(2), 255
    
  Case 41 'set.view(oA, B)
    'validate that view exists
    If Not agViews.Exists(ArgVal(1)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4124), ARG1, CStr(ArgVal(1)))
        Exit Function
      Case leMedium
        AddWarning LoadResString(5037)
      Case leLow
      End Select
    End If
  
  Case 49 To 53, 97, 118  'last.cel, current.cel, current.loop,
                          'current.view, number.of.loops, get.room.v
                          'get.num
    'variable arg is second
    CheckResVarUse ArgVal(1), 255
  
  Case 54 'set.priority(oA, B)
    'check priority value
    If ArgVal(1) > 15 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4125)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5050)
      Case leLow
      End Select
    End If
    
  Case 57 'get.priority
    'variable is second argument
    'assume 15 (worst case scenario)
    CheckResVarUse ArgVal(1), 15
  
  Case 63 'set.horizon(A)
    'validate horizon value
    Select Case agMainLogSettings.ErrorLevel
    Case leHigh
      If ArgVal(0) > 167 Then
        strErrMsg = LoadResString(4126)
        Exit Function
      ElseIf ArgVal(0) > 120 Then
        AddWarning LoadResString(5042)
      End If
    Case leMedium
      If ArgVal(0) > 167 Then
        AddWarning LoadResString(5043)
      ElseIf ArgVal(0) > 120 Then
        AddWarning LoadResString(5042)
      End If
    Case leLow
    End Select
    
  Case 69 'distance
    'variable is third arg
    CheckResVarUse ArgVal(2), 255
  
  Case 73, 75, 99 'end.of.loop, reverse.loop
    'check for reserved flags
    CheckResFlagUse ArgVal(1)
    
  Case 81 'move.obj(oA, B,C,D,fDONE)
    'validate the target position
    If ArgVal(1) > 159 Or ArgVal(2) > 167 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4128)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5062)
      Case leLow
      End Select
    End If
    
    'check for ego object
    If ArgVal(0) = 0 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5045)
      Case leLow
      End Select
    End If
    
    'check for use of reserved flags
    CheckResFlagUse ArgVal(4)
  
  Case 82 'move.obj.v
    'check for use of reserved flags
    CheckResFlagUse ArgVal(4)
  
  Case 83 'follow.ego(oA, B, fDONE)
    'check for ego object
    If ArgVal(0) = 0 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5027)
      Case leLow
      End Select
    End If
    
    'check for use of reserved flags
    CheckResFlagUse ArgVal(2)
  
  Case 86 'set.dir(oA, vB)
    'check for ego object
    If ArgVal(0) = 0 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5026)
      Case leLow
      End Select
    End If
    
  Case 87 'get.dir
    'variable is second arg
    'assume 8 (worst case scenario)
    CheckResVarUse ArgVal(1), 8
    
  Case 90 'block(x1,y1,x2,y2)
    'validate that all are within bounds, and that x1<=x2 and y1<=y2
    If ArgVal(0) > 159 Or ArgVal(1) > 167 Or ArgVal(2) > 159 Or ArgVal(3) > 167 Then
      'bad number
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4129)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5020)
      Case leLow
      End Select
    End If
    
    If (ArgVal(0) > ArgVal(2)) Or (ArgVal(1) > ArgVal(3)) Then
      'can't work
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4129)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5051)
      Case leLow
      End Select
    End If
    
  Case 98 'load.sound(A)
    'validate the sound exists
    If Not agSnds.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4130), ARG1, CStr(ArgVal(0)))
        Exit Function
      Case leMedium
        AddWarning LoadResString(5014)
      Case leLow
      End Select
    End If
    
  Case 103 'display(ROW,COL,mC)
    'check row/col against limits
    If ArgVal(0) > 24 Or ArgVal(1) > 39 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
       strErrMsg = LoadResString(4131)
       Exit Function
      Case leMedium
        AddWarning LoadResString(5059)
      Case leLow
      End Select
    End If
    
  Case 105 'clear.lines(TOP,BTM,C)
    'top must be >btm; both must be >24
    If ArgVal(0) > 24 Or ArgVal(1) > 24 Or ArgVal(0) > ArgVal(1) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4132)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5011)
      Case leLow
      End Select
    End If
    
  Case 109 'set.text.attribute(A,B)
    'should be limited to valid color values (0-15)
    If ArgVal(0) > 15 Or ArgVal(1) > 15 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4133)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5029)
      Case leLow
      End Select
    End If
    
  Case 110 'shake.screen(A)
    'shouldn't normally have more than a few shakes; zero is BAD
    If ArgVal(0) = 0 Then
      'error!
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        strErrMsg = LoadResString(4134)
        Exit Function
      Case leLow
      End Select
    ElseIf ArgVal(0) > 15 Then
      'could be a palette change?
      If ArgVal(0) >= 100 And ArgVal(0) <= 109 Then
        'separate warning
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh, leMedium
          AddWarning LoadResString(5058)
        Case leLow
        End Select
      Else
        'warning
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh, leMedium
          AddWarning LoadResString(5057)
        Case leLow
        End Select
      End If
    End If
  
  Case 111 'configure.screen(TOP,INPUT,STATUS)
    'top should be <=3
    'input and status should not be equal
    'input and status should be <top or >=top+21
    If ArgVal(0) > 3 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4135)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5044)
      Case leLow
      End Select
    End If
    If ArgVal(1) = ArgVal(2) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leLow
        AddWarning LoadResString(5048)
      Case leLow
      End Select
    End If
    If (ArgVal(1) >= ArgVal(0) And ArgVal(1) <= ArgVal(0) + 20) Or (ArgVal(2) >= ArgVal(0) And ArgVal(2) <= ArgVal(0) + 20) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5049)
      Case leLow
      End Select
    End If
    
  Case 115 'get.string(sA, mB, ROW,COL,LEN)
    'if row>24, both row/col are ignored; len is limited automatically to <=40
    If ArgVal(2) > 24 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5052)
      Case leLow
      End Select
    End If
    If ArgVal(4) > 40 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5056)
      Case leLow
      End Select
    End If
    
  Case 121 'set.key(A,B,cC)
    'only 49 assignments can  be made
    'movement keys not allowed to be assigned
    'ctrlr value limited to 0-49
    
    'increment controller count
    intCtlCount = intCtlCount + 1
    
    'if exceeds max
    If intCtlCount >= 50 Then
      'too many
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4137)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5016)
      Case leLow
      End Select
    End If
    
    'validate argument
    If ArgVal(2) > 49 Then
      'invalid number
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4154)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5065)
      Case leLow
      End Select
    End If
    
    'check for improper key assignments
    If ArgVal(1) = 0 Then
      Select Case ArgVal(0)
      Case 8, 13, 32, 71, 72, 73, 75, 76, 77, 79, 80, 81, 82, 83
        'bad
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          strErrMsg = LoadResString(4155)
          Exit Function
        Case leMedium
          AddWarning LoadResString(5066)
        Case leLow
        End Select
      End Select
    End If
    
  Case 122 'add.to.pic(VIEW,LOOP,CEL,X,Y,PRI,MGN)
    'VIEW, LOOP & CEL must exist
    'CEL width must be >=3
    'x,y must be within limits
    'PRI must be <=15
    'MGN must be 0-3
    
    'validate view
    If Not agViews.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4138), ARG1, CStr(ArgVal(0)))
        Exit Function
      Case leMedium
        AddWarning Replace(LoadResString(5064), ARG1, "view")
      Case leLow
      End Select
    End If
    'load view to test loop & cel
    Dim blnUnLoad As Boolean
    blnUnLoad = Not agViews(ArgVal(0)).Loaded
    Err.Clear
    If blnUnLoad Then
      On Error Resume Next
      agViews(ArgVal(0)).Load
    End If
    If Err.Number = 0 Then
      'validate loop
      If ArgVal(1) >= agViews(ArgVal(0)).Loops.Count Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          strErrMsg = Replace(Replace(LoadResString(4139), ARG1, CStr(ArgVal(1))), ARG2, CStr(ArgVal(0)))
          If blnUnLoad Then
            agViews(ArgVal(0)).Unload
          End If
          Exit Function
        Case leMedium
          AddWarning Replace(LoadResString(5065), ARG1, "loop")
        Case leLow
        End Select
      End If
      'validate cel
      If ArgVal(2) >= agViews(ArgVal(0)).Loops(ArgVal(1)).Cels.Count Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          strErrMsg = Replace(Replace(Replace(LoadResString(4140), ARG1, CStr(ArgVal(2))), ARG2, CStr(ArgVal(1))), ARG3, CStr(ArgVal(0)))
          If blnUnLoad Then
            agViews(ArgVal(0)).Unload
          End If
          Exit Function
        Case leMedium
          AddWarning Replace(LoadResString(5064), ARG1, "cel")
        Case leLow
        End Select
      End If
    Else
      'can't load the view; add a warning
      Err.Clear
      AddWarning Replace(LoadResString(5021), ARG1, CStr(ArgVal(0)))
    End If
    If blnUnLoad Then
      agViews(ArgVal(0)).Unload
    End If
    
    On Error GoTo ErrHandler
    
    'x,y must be within limits
    If ArgVal(3) > 159 Or ArgVal(4) > 167 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4141)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5038)
      Case leLow
      End Select
    End If
    
    'PRI must be <=15
    If ArgVal(5) > 15 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4142)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5050)
      Case leLow
      End Select
    End If
    'MGN must be 0-3
    If ArgVal(6) > 15 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4143)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5041)
      Case leLow
      End Select
    End If
  
  Case 129 'show.obj(VIEW)
    'validate view
    If Not agViews.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4144), ARG1, CStr(ArgVal(0)))
        Exit Function
      Case leMedium
        AddWarning LoadResString(5061)
      Case leLow
      End Select
    End If
    
  Case 130 'random(LOWER,UPPER,vC)
    'lower should be <= upper
    If ArgVal(0) > ArgVal(1) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4145)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5054)
      End Select
    End If
    'check for use of reserved variables (use upper bound as worst-case scenario for value)
    CheckResVarUse ArgVal(2), ArgVal(1)
  
  Case 142 'script.size
    'raise warning/error if in other than logic0
    If bytLogComp <> 0 Then
      'warn
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        'set warning
        AddWarning LoadResString(5039)
      Case leLow
        'no action
      End Select
    End If
    'check for absurdly low value for script size
    If ArgVal(0) < 10 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5009)
      Case leLow
      End Select
    End If
        
  Case 147 'reposition.to(oA, B,C)
    'validate the target position
    If ArgVal(1) > 159 Or ArgVal(2) > 167 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4128)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5062)
      Case leLow
      End Select
    End If
  
  Case 150 'trace.info(LOGIC,ROW,HEIGHT)
    'logic must exist
    'row + height must be <22
    'height must be >=2 (but interpreter checks for this error)
    
    'validate that logic exists
    If Not agLogs.Exists(ArgVal(0)) Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = Replace(LoadResString(4153), ARG1, CStr(ArgVal(0)))
        Exit Function
      Case leMedium
        AddWarning LoadResString(5040)
      Case leLow
      End Select
    End If
    'validate that height is not too small
    If ArgVal(2) < 2 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5046)
      Case leLow
      End Select
    End If
    'validate size of window
    If ArgVal(1) + ArgVal(2) > 22 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4146)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5063)
      Case leLow
      End Select
    End If
    
  Case 151, 152 'Print.at(mA, ROW, COL, MAXWIDTH), print.at.v
    'row>=1, <=22
    'col >=2
    'maxwidth <=36
    'col + maxwidth <=39
    If ArgVal(1) = 0 Or ArgVal(1) > 22 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4147)
        Exit Function
        
      Case leMedium
        AddWarning LoadResString(5067)
      Case leLow
      End Select
    End If
    
    If ArgVal(2) < 2 Or ArgVal(2) + ArgVal(3) > 39 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4148)
        Exit Function
        
      Case leMedium
        AddWarning LoadResString(5068)
      Case leLow
      End Select
    End If
    
  Case 154 'clear.text.rect(R1,C1,R2,C2,COLOR)
    'If either row argument is >24,
    'or either column argument is >39,
    'or R2 < R1 or C2 < C1,
    'the results are unpredictable
    If ArgVal(0) > 24 Or ArgVal(1) > 39 Or _
       ArgVal(2) > 24 Or ArgVal(3) > 39 Or _
       ArgVal(2) < ArgVal(0) Or ArgVal(3) < ArgVal(1) Then
      'invalid items
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4150)
        Exit Function
      Case leMedium
        'if due to pos2 < pos1
        If ArgVal(2) < ArgVal(0) Or ArgVal(3) < ArgVal(1) Then
          AddWarning LoadResString(5069)
        End If
        'if due to variables outside limits
        If ArgVal(0) > 24 Or ArgVal(1) > 39 Or _
           ArgVal(2) > 24 Or ArgVal(3) > 39 Then
          AddWarning LoadResString(5070)
        End If
      End Select
    End If
    
    'check color
    If ArgVal(4) > 15 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4133)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5029)
      Case leLow
      End Select
    End If
    
  Case 158 'submit.menu()
    'should only be called in logic0
    'raise warning/error if in other than logic0
    If bytLogComp <> 0 Then
      'warn
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        'set warning
        AddWarning LoadResString(5047)
      Case leLow
      End Select
    End If
    
  Case 174 'set.pri.base(A)
    'calling set.pri.base with value >167 doesn't make sense
    If ArgVal(0) > 167 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh, leMedium
        AddWarning LoadResString(5071)
      Case leLow
      End Select
    End If
  End Select

  'success
  ValidateArgs = True
Exit Function

ErrHandler:
  Debug.Assert False
  Resume Next
End Function

Private Function ValidateIfArgs(ByVal CmdNum As Long, ByRef ArgVal() As Byte) As Boolean

  'check for specific command issues
  
  Select Case CmdNum
  Case 9 'has (iA)
  Case 10 'obj.in.room(iA, vB)
  Case 11, 16, 17, 18 'posn(oA, X1,Y1,X2,Y2)
                      'obj.in.box(oA, B,C,D,E)
                      'center.posn(oA, B,C,D,E)
                      'right.posn(oA, B,C,D,E)
    'validate that all are within bounds, and that x1<=x2 and y1<=y2
    If ArgVal(1) > 159 Or ArgVal(2) > 167 Or ArgVal(3) > 159 Or ArgVal(4) > 167 Then
      'bad number
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4151)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5072)
      Case leLow
      End Select
    End If
    
    If (ArgVal(1) > ArgVal(3)) Or (ArgVal(2) > ArgVal(4)) Then
      'can't work
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        strErrMsg = LoadResString(4151)
        Exit Function
      Case leMedium
        AddWarning LoadResString(5073)
      Case leLow
      End Select
    End If
  
  Case 12 'controller (cA)
    'has controller been assigned?
    'not sure how to check it; calls to controller cmd may
    'occur in logics that are compiled before the logic that sets
    'them up...
    
  Case 14 'said()
  Case 15 'compare.strings(sA, sB)
  
  End Select

  'success
  ValidateIfArgs = True
End Function

Private Function VariableValue(strVar As String) As Integer
  'this function will extract the variable number from
  'an input variable string
  'the input string should be of the form #, a# or *a#
  ' where a is a valid variable prefix (v, f, s, m, w, c)
  'and # is 0-255
  'if the result is invalid, this function returns -1
  
  Dim strVarVal As String
  Dim intVarVal  As Integer
  Dim blnOutofBounds As Boolean
  
  On Error GoTo ErrHandler
  
  'if not numeric
  If Not IsNumeric(strVar) Then
    'strip off variable prefix, and indirection
    'if indirection
    If Left$(strVar, 1) = "*" Then
      strVarVal = Right$(strVar, Len(strVar) - 2)
    Else
      strVarVal = Right$(strVar, Len(strVar) - 1)
    End If
  Else
    'use the input value
    strVarVal = strVar
  End If
  
  'if result is a number
  If IsNumeric(strVarVal) Then
    'get number
    intVarVal = Val(strVarVal)
    'for word only, subtract one to
    'account for '1' based word data type
    '(i.e. w1 is first word, but command uses arg value of '0')
    If AscW(strVar) = 119 Then
      intVarVal = intVarVal - 1
    End If
    
    'verify within bounds  0-255
    If intVarVal >= 0 And intVarVal <= 255 Then
      'return this value
      VariableValue = intVarVal
      Exit Function
    End If
  End If
  
  'error- return -1
  VariableValue = -1
Exit Function

ErrHandler:
  Err.Clear
  'return -1
  VariableValue = -1
End Function
Private Function WriteMsgs() As Boolean
  'this function will write the messages for a logic at the end of
  'the resource.
  'messages are encrypted with the string 'Avis Durgan'. No gaps
  'are allowed, so messages that are skipped must be included as
  'zero length messages

  Dim lngMsgSecStart As Long
  Dim lngMsgSecLen As Long
  Dim lngMsgPos(255) As Long
  Dim intCharPos As Integer, bytCharVal As Byte
  Dim lngMsg As Long
  Dim lngMsgCount As Integer
  Dim lngCryptStart As Long
  Dim lngMsgLen As Integer
  Dim i As Integer
  
  On Error GoTo ErrHandler
  
  'calculate start of message section
  lngMsgSecStart = tmpLogRes.Size
  
  'find last message by counting backwards until a msg is found
  lngMsgCount = 256
  Do
    lngMsgCount = lngMsgCount - 1
  Loop Until blnMsg(lngMsgCount) Or (lngMsgCount = 0)
  
  'write msg count,
  tmpLogRes.WriteByte CByte(lngMsgCount)
  
  'write place holder for msg end
  tmpLogRes.WriteWord 0&
  'write place holders for msg pointers
  For i = 1 To lngMsgCount
    tmpLogRes.WriteWord 0&
  Next i
  
  'begin encryption process
  lngCryptStart = tmpLogRes.Size
  For lngMsg = 1 To lngMsgCount
    If blnMsg(lngMsg) Then
      'check for null msg
      If strMsg(lngMsg) = NULL_MSG Then
        strMsg(lngMsg) = vbNullString
      End If
      'get length
      lngMsgLen = Len(strMsg(lngMsg))
      'calculate offset to start of this message (adjust by one byte, which
      'is the byte that indicates how many msgs there are)
      lngMsgPos(lngMsg) = tmpLogRes.GetPos - (lngMsgSecStart + 1)
      If lngMsgLen > 0 Then
        'step through all characters in this msg
        intCharPos = 1
        Do
          'get ascii code for this character
          bytCharVal = Asc(Mid$(strMsg(lngMsg), intCharPos))
          'check for special codes ("\" = 92
          If (bytCharVal = 92) And (intCharPos < lngMsgLen) Then
            'check against string value so "n" and "N" are treated
            'the same (since Option Compare Text applies to this module)
            If Mid$(strMsg(lngMsg), intCharPos + 1, 1) = "n" Then
              ' \n = new line
              bytCharVal = &HA
              intCharPos = intCharPos + 1
            'here we check the ascii value of the string, since this isn't an alphabetic character
            ElseIf AscW(Mid$(strMsg(lngMsg), intCharPos + 1)) = 34 Then 'QUOTECHAR
              '\" = quote mark (chr$(34))
              bytCharVal = &H22
              intCharPos = intCharPos + 1
            'here we check the ascii value of the string, since this isn't an alphabetic character
            ElseIf AscW(Mid$(strMsg(lngMsg), intCharPos + 1)) = 92 Then ' = "\"
              '\\ = \
              bytCharVal = &H5C
              intCharPos = intCharPos + 1
            End If
          End If
          
          'write the encrypted byte (need to adjust for previous
          'messages, and current position)
          tmpLogRes.WriteByte bytCharVal Xor bytEncryptKey((tmpLogRes.GetPos - lngCryptStart) Mod 11)
          
          'increment pointer
          intCharPos = intCharPos + 1
        Loop Until intCharPos > Len(strMsg(lngMsg))
      End If
      'add a trailing zero to terminate message
      tmpLogRes.WriteByte &H0 Xor bytEncryptKey((tmpLogRes.GetPos - lngCryptStart) Mod 11)
    Else
      'if msg does not exist, set its position to zero
      lngMsgPos(lngMsg) = 0
    End If
  Next lngMsg
  
  'calculate length of msg section, and write it at beginning
  'of msg section (adjust by one byte, which is the
  'byte that indicates number of msgs written)
  lngMsgSecLen = tmpLogRes.GetPos - (lngMsgSecStart + 1)
  tmpLogRes.WriteWord lngMsgSecLen, lngMsgSecStart + 1
  
  'write msg section start value at beginning of resource
  '(correct by two so it gives position relative to byte 7 of
  'the logic resource header - see procedure 'DecodeLogic' for details)
  tmpLogRes.WriteWord lngMsgSecStart - 2, 0

  'write all the msg pointers
  '
  For lngMsg = 1 To lngMsgCount
    tmpLogRes.WriteWord lngMsgPos(lngMsg), lngMsgSecStart + 1 + lngMsg * 2
  Next lngMsg
  
  'and return true
  WriteMsgs = True
  
Exit Function

ErrHandler:
  'any errors, means there is a problem
  blnError = True
  Resume Next
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG2, "WriteMsgs")
  Err.Clear
  WriteMsgs = False
End Function



Private Function ReadLabels() As Boolean

  Dim i As Byte
  Dim intLabel As Integer
  Dim strLabel As String
  Dim strLineSplit() As String
  Dim rtn As Long
  
  On Error GoTo ErrHandler
  
  'reset counter
  bytLabelCount = 0
  
  'reset compiler to first input line
  ResetCompiler
  
  Do
    'look for label name
    If InStrB(1, strCurrentLine, ":") <> 0 Then
      strLineSplit = Split(strCurrentLine, ":")
      'if there are exactly two lines split,
      If UBound(strLineSplit) = 1 Then
        'second should be empty (or just spaces)
        If LenB(Trim$(strLineSplit(1))) = 0 Then
          'first should be label name,
          strLabel = Trim$(strLineSplit(0))
          'make sure enough room
          If bytLabelCount >= MaxLabels Then
            strErrMsg = Replace(LoadResString(4109), ARG1, CStr(MaxLabels))
            Exit Function
          End If
          
          rtn = ValidateDefName(strLabel)
          'numbers are ok for labels
          If rtn = 2 Then
            rtn = 0
          End If
          If rtn <> 0 Then
            'error
            Select Case rtn
            Case 1
              strErrMsg = LoadResString(4096)
            Case 3
              strErrMsg = Replace(LoadResString(4025), ARG1, strLabel)
            Case 4
              strErrMsg = Replace(LoadResString(4026), ARG1, strLabel)
            Case 5
              strErrMsg = Replace(LoadResString(4028), ARG1, strLabel)
            Case 6
              strErrMsg = LoadResString(4091)
            Case 7
              strErrMsg = Replace(LoadResString(4024), ARG1, strLabel)
            Case 8
              strErrMsg = Replace(LoadResString(4033), ARG1, strLabel)
            Case 9
              strErrMsg = Replace(LoadResString(4030), ARG1, strLabel)
            Case 10
              strErrMsg = Replace(LoadResString(4029), ARG1, strLabel)
            Case 11
              strErrMsg = Replace(LoadResString(4032), ARG1, strLabel)
            Case 12
              strErrMsg = Replace(LoadResString(4031), ARG1, strLabel)
            Case 13
              strErrMsg = LoadResString(4068)
            End Select
            Exit Function
          End If
          
          'no periods allowed either
          If InStrB(1, strLabel, ".") <> 0 Then
            strErrMsg = LoadResString(4068)
            Exit Function
          End If
            
          'check label against current list of labels
          If bytLabelCount > 0 Then
            For i = 1 To bytLabelCount
              If strLabel = llLabel(i).Name Then
                strErrMsg = Replace(LoadResString(4027), ARG1, strLabel)
                Exit Function
              End If
            Next i
          End If

          'increment number of labels, and save
          bytLabelCount = bytLabelCount + 1
          llLabel(bytLabelCount).Name = strLabel
          llLabel(bytLabelCount).Loc = 0
        End If
      End If
    End If
    
    'get next line
    IncrementLine
  Loop Until lngLine = -1
  ReadLabels = True
Exit Function
ErrHandler:
  strErrMsg = Replace(Replace(LoadResString(4115), ARG1, CStr(Err.Number)), ARG1, "ReadLabels")
  Err.Clear
End Function
Private Function CompileAGI() As Boolean
  'main compiler function for non-VB input
  'steps through input one command at a time and converts it
  'to AGI logic code
  'Note that don't need to set blnError flag here;
  'an error will cause this function to return a value of false
  'which causes the compiler to display error info
  
  Const MaxGotos = 255

  Dim strNextCmd As String
  Dim strPrevCmd As String
  Dim strArg As String, bytArg(7) As Byte
  Dim i As Integer
  Dim intCmdNum As Integer
  Dim BlockStartDataLoc(MaxBlockDepth) As Integer
  Dim BlockDepth As Integer
  Dim BlockIsIf(MaxBlockDepth) As Boolean
  Dim BlockLength(MaxBlockDepth) As Integer
  Dim CurLabel As Integer
  Dim intLabelNum As Integer
  Dim Gotos(MaxGotos) As LogicGoto
  Dim NumGotos As Integer
  Dim GotoData As Long
  Dim CurGoto As Integer
  Dim blnLastCmdRtn As Boolean
  Dim lngReturnLine As Long
  
  On Error GoTo ErrHandler
  
  'initialize variables
  BlockDepth = 0
  NumGotos = 0
  'reset compiler
  ResetCompiler
  
  blnError = False
  
  'get first command
  strNextCmd = NextCommand()
    
  'process commands in the input string list until finished
  Do Until lngLine = -1
    'reset last command flag
    blnLastCmdRtn = False
    lngReturnLine = 0
    
    'process the command
    Select Case strNextCmd
    Case "{"
      'can't have a "{" command, unless it follows an 'if' or 'else'
      If strPrevCmd <> "if" And strPrevCmd <> "else" Then
        strErrMsg = LoadResString(4008)
        Exit Function
      End If
      
    Case "}"
      'if no block currently open,
      If BlockDepth = 0 Then
        strErrMsg = LoadResString(4010)
        Exit Function
      End If
      'if last position in resource is two bytes from start of block
      If tmpLogRes.Size = BlockStartDataLoc(BlockDepth) + 2 Then
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          strErrMsg = LoadResString(4049)
          Exit Function
        Case leMedium
          'set warning
          AddWarning LoadResString(5001)
        Case leLow
          'no action
        End Select
      End If
      'calculate and write block length
      BlockLength(BlockDepth) = tmpLogRes.Size - BlockStartDataLoc(BlockDepth) - 2
      tmpLogRes.WriteWord CLng(BlockLength(BlockDepth)), CLng(BlockStartDataLoc(BlockDepth))
      'remove block from stack
      BlockDepth = BlockDepth - 1
    Case "if"
      'compile the 'if' statement
      If Not CompileIf() Then
        Exit Function
      End If
      'if block stack exceeded
      If BlockDepth >= MaxBlockDepth Then
        strErrMsg = Replace(LoadResString(4110), ARG1, CStr(MaxBlockDepth))
        Exit Function
      End If
      'add block to stack
      BlockDepth = BlockDepth + 1
      BlockStartDataLoc(BlockDepth) = tmpLogRes.GetPos
      BlockIsIf(BlockDepth) = True
      'write placeholders for block length
      tmpLogRes.WriteWord &H0

    Case "else"
      'else can only follow a close bracket
      If strPrevCmd <> "}" Then
       
        strErrMsg = LoadResString(4011)
        Exit Function
      End If
      
      'if the block closed by that bracket was an 'else'
      '(which will be determined by having that block's IsIf flag NOT being set),
      If Not BlockIsIf(BlockDepth + 1) Then
        strErrMsg = LoadResString(4083)
        Exit Function
      End If
      
      'adjust blockdepth to the 'if' command
      'directly before this 'else'
      BlockDepth = BlockDepth + 1
      'adjust previous block length to accomodate the 'else' statement
      BlockLength(BlockDepth) = BlockLength(BlockDepth) + 3
      tmpLogRes.WriteWord CLng(BlockLength(BlockDepth)), CLng(BlockStartDataLoc(BlockDepth))
      'previous 'if' block is now closed; use same block level
      'for this 'else' block
      BlockIsIf(BlockDepth) = False
      'write the 'else' code
      tmpLogRes.WriteByte &HFE
      BlockStartDataLoc(BlockDepth) = tmpLogRes.GetPos
      tmpLogRes.WriteWord &H0  ' block length filled in later.
      
    Case "goto"
        'next command should be "("
        If NextChar() <> "(" Then
          strErrMsg = LoadResString(4001)
          Exit Function
        End If
        'get goto argument
        strArg = NextCommand()
        
        'if argument is NOT a valid label
        If LabelNum(strArg) = 0 Then
          strErrMsg = Replace(LoadResString(4074), ARG1, strArg)
          Exit Function
        End If
        'if too many gotos
        If NumGotos >= MaxGotos Then
          strErrMsg = Replace(LoadResString(4108), ARG1, CStr(MaxGotos))
        End If
        'save this goto info on goto stack
        NumGotos = NumGotos + 1
        Gotos(NumGotos).LabelNum = LabelNum(strArg)
        'write goto command byte
        tmpLogRes.WriteByte &HFE
        Gotos(NumGotos).DataLoc = tmpLogRes.GetPos
        'write placeholder for amount of offset
        tmpLogRes.WriteWord &H0
        'next character should be ")"
        If NextChar() <> ")" Then
          strErrMsg = LoadResString(4003)
          Exit Function
        End If
        'verify next command is end of line (;)
        If NextChar() <> ";" Then
          blnError = True
          strErrMsg = LoadResString(4007)
          Exit Function
        End If
   
    Case Else
      'must be a label, command, or special syntax
      'if next character is a colon
      If Mid$(strCurrentLine, lngPos + 1, 1) = ":" Then
        'it's a label
        intLabelNum = LabelNum(strNextCmd)
        'if not a valid label
        If intLabelNum = 0 Then
          strErrMsg = LoadResString(4076)
          Exit Function
        End If
        'save position of label
        llLabel(intLabelNum).Loc = tmpLogRes.Size
        'read in next char to skip past the colon
        NextChar
      Else
        'get number of command
        intCmdNum = CommandNum(False, strNextCmd)
        'if invalid version
        If intCmdNum = 254 Then
          'raise error
          strErrMsg = Replace(LoadResString(4065), ARG1, strNextCmd)
          Exit Function
        'if command not found,
        ElseIf intCmdNum = 255 Then  ' not found
          'try to parse special syntax
          If CompileSpecial(strNextCmd) Then
            'check for error
            If blnError Then
              Exit Function
            End If
          Else
            'unknown command
            strErrMsg = Replace(LoadResString(4116), ARG1, strNextCmd)
            Exit Function
          End If
        Else
          'write the command code,
          tmpLogRes.WriteByte CByte(intCmdNum)
          'next character should be "("
          If NextChar() <> "(" Then
            strErrMsg = LoadResString(4048)
            Exit Function
          End If
          
          'now extract arguments,
          For i = 0 To agCmds(CByte(intCmdNum)).ArgCount - 1
            'after first argument, verify comma separates arguments
            If i > 0 Then
              If NextChar(True) <> "," Then
                strErrMsg = Replace(LoadResString(4047), ARG1, CStr(i))
                Exit Function
              End If
            End If
            bytArg(i) = GetNextArg(agCmds(CByte(intCmdNum)).ArgType(i), i)
            'if error
            If blnError Then
              Exit Function
            End If
            
            'write argument
            tmpLogRes.WriteByte bytArg(i)
          Next i
          
          'validate arguments for this command
          If Not ValidateArgs(intCmdNum, bytArg()) Then
            Exit Function
          End If
          
          'next character must be ")"
          If NextChar() <> ")" Then
            strErrMsg = LoadResString(4048)
            Exit Function
          End If
          If intCmdNum = 0 Then
            blnLastCmdRtn = True
            'set line number
            If lngReturnLine = 0 Then
              lngReturnLine = lngLine + 1
            End If
          End If
        End If
        
        'verify next command is end of line (;)
        If NextChar(True) <> ";" Then
          strErrMsg = LoadResString(4007)
          Exit Function
        End If
      End If
     End Select
    'get next command
    strPrevCmd = strNextCmd
    strNextCmd = NextCommand()
 Loop
    
  If (Not blnLastCmdRtn) Then
    strErrMsg = LoadResString(4102)
    Exit Function
  End If
  
  'check to see if everything was wrapped up properly
  If BlockDepth > 0 Then
    strErrMsg = LoadResString(4009)
    'reset errorline to return cmd
    lngErrLine = lngReturnLine
    Exit Function
  End If
  
  'write in goto values
  For CurGoto = 1 To NumGotos
    GotoData = llLabel(Gotos(CurGoto).LabelNum).Loc - Gotos(CurGoto).DataLoc - 2
    If GotoData < 0 Then
      'need to convert it to an unsigned integer value
      GotoData = &H10000 + GotoData
    End If
    tmpLogRes.WriteWord CLng(GotoData), CLng(Gotos(CurGoto).DataLoc)
  Next CurGoto
  
  'return true
  CompileAGI = True
Exit Function

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 658, strErrSrc, Replace(LoadResString(658), ARG1, CStr(lngError) & ":" & strError)
End Function

Private Function MessageNum(strMsgIn As String) As Byte
  ' Returns the number of the message corresponding to
  'strMsg, or creates a new msg number if strMsg is not
  'currently a message
  'if maximum number of msgs assigned, returns  0
  Dim intMsg As Integer
  
  'blank msgs not allowed
  If LenB(strMsgIn) = 0 Then
    Select Case agMainLogSettings.ErrorLevel
    Case leHigh, leMedium
      AddWarning LoadResString(5074)
    Case leLow
      'allow it
    End Select
    
    'use special character string 255-255 as place holder
    strMsgIn = NULL_MSG
  End If
  
  For intMsg = 1 To 255
    'if this is the message
    If strMsg(intMsg) = strMsgIn Then
      'return this value
      MessageNum = CByte(intMsg)
      Exit Function
    End If
  Next intMsg
  
  'msg doesn't exist; find an empty spot
  For intMsg = 1 To 255
    If Not blnMsg(intMsg) Then
      'this message is available
      blnMsg(intMsg) = True
      strMsg(intMsg) = strMsgIn
      MessageNum = CByte(intMsg)
      Exit Function
    End If
  Next intMsg
  'if no room found, return zero
  MessageNum = 0
End Function

Private Function CommandNum(blnIF As Boolean, strCmdName As String) As Byte
  'gets the command number
  'of a command, based on the text

  If blnIF Then
    For CommandNum = 1 To agNumTestCmds
      If strCmdName = agTestCmds(CommandNum).Name Then
        Exit Function
      End If
    Next CommandNum
  Else
    For CommandNum = 0 To agNumCmds
      If strCmdName = agCmds(CommandNum).Name Then
        Exit Function
      End If
    Next CommandNum
    'maybe the command is a valid agi command, but
    'just not supported in this agi version
    'If agNumCmds <> 182 Then
      For CommandNum = agNumCmds + 1 To 182
        If strCmdName = agCmds(CommandNum).Name Then
          'set value so compiler knows this is valid cmd,
          'but invalid for this version
          CommandNum = 254
          Exit Function
        End If
      Next CommandNum
    'end if
  End If

  CommandNum = 255
End Function

Private Function CompileSpecialIf(strArg1 As String, blnNOT As Boolean) As Boolean
  'this funtion determines if strArg1 is a properly
  'formatted special test syntax
  'and writes the appropriate data to the resource
  
  '(proprerly formatted special IF syntax will be one of following:
  ' v## expr v##
  ' v## expr ##
  ' f##
  '
  ' where ## is a number from 1-255
  ' and expr is '"==", "!=", "<", ">=", "=>", ">", "<=", "=<"
  ' (for VB syntax, '==' is replaced by '=', and '!=' is replaced by '<>')
  
  Dim strArg2 As String
  Dim intArg1 As Integer, intArg2 As Integer
  Dim blnArg2Var As Boolean
  Dim blnAddNOT As Boolean
  Dim bytCmdNum As Byte
  
  'check for variable argument
  If Not ConvertArgument(strArg1, atVar) Then
    'check for flag argument
    If Not ConvertArgument(strArg1, atFlag) Then
      'invalid argument
      blnError = True
      strErrMsg = LoadResString(4039)
      Exit Function
    End If
  End If
  
  'arg in can only be f# or v#
  Select Case Left$(strArg1, 1)
  Case "f"
    'get flag argument value
    intArg1 = VariableValue(strArg1)
    'if invalid flag number
    If intArg1 = -1 Then
      'invalid number
        blnError = True
        strErrMsg = Replace(LoadResString(4066), ARG1, "1")
      Exit Function
    End If
    'write isset cmd
    tmpLogRes.WriteByte &H7  ' isset
    tmpLogRes.WriteByte CByte(intArg1)
  
  Case "v"
    'can't have a NOT in front of variable comparisons
    If blnNOT Then
      blnError = True
      strErrMsg = LoadResString(4098)
      Exit Function
    End If
    
    'arg 1 must be 'v#' format
    intArg1 = VariableValue(strArg1)
    
    'if invalid variable number
    If intArg1 = -1 Then
      'invalid number
        blnError = True
        strErrMsg = LoadResString(4086)
      Exit Function
    End If
    
    'get comparison expression
    strArg2 = NextCommand()
    'get command code for this expression
    Select Case strArg2
    Case strEqual
      bytCmdNum = &H1
    Case strNotEqual
      bytCmdNum = &H1
      blnAddNOT = True
    Case ">"
      bytCmdNum = &H5
    Case "<=", "=<"
      bytCmdNum = &H5
      blnAddNOT = True
    Case "<"
      bytCmdNum = &H3
    Case ">=", "=>"
      bytCmdNum = &H3
      blnAddNOT = True
    Case Else
      blnError = True
      strErrMsg = LoadResString(4078)
      Exit Function
    End Select
         
    'get second argument (numerical or variable)
    blnArg2Var = True
    intArg2 = GetNextArg(atNum, -1, blnArg2Var)
    'if error
    If blnError Then
      'reset error msg
      strErrMsg = LoadResString(4089)
      Exit Function
    End If
     
    'if comparing to a variable,
    If blnArg2Var Then
      bytCmdNum = bytCmdNum + 1
    End If
    
    'if adding a 'not'
    If blnAddNOT Then
      tmpLogRes.WriteByte (&HFD)
    End If
    
    'write command, and arguments
    tmpLogRes.WriteByte bytCmdNum
    tmpLogRes.WriteByte CByte(intArg1)
    tmpLogRes.WriteByte CByte(intArg2)
  End Select
    
  'return true
  CompileSpecialIf = True
End Function

Private Function CompileSpecial(strArgIn As String) As Boolean
  'strArg1 should be a variable in the format of *v##, v## or s##
  'if it is not, this function will trap it and return an error
  'the expression after strArg1 should be one of the following:
  ' =, +=, -=, *=, /=, ++, --
  '
  'after determining assignment type, this function will validate additional
  'arguments as necessary
  '
  'Note: in VB syntax, unary operators (++, --)  and short hand expressions
  '(+=, -=, *=, /=) are not allowed
  
  Dim strArg1 As String, strArg2 As String
  Dim intArg1 As Integer, intArg2 As Integer
  Dim blnArg2Var  As Boolean
  Dim intDir As Integer '0 = no indirection; 1 = left; 2 = right
  Dim bytCmd As Byte
  On Error GoTo ErrHandler
  
  strArg1 = strArgIn
  
  'assume this is a special syntax until proven otherwise
  CompileSpecial = True
  
  'if this is indirection
  If Left$(strArg1, 1) = "*" Then
    'left indirection
    '     *v# = #
    '     *v# = v#
    
    'next char can't be a space, newline, or tab
    Select Case Mid$(strCurrentLine, lngPos + 1, 1)
    Case " ", vbTab, vbNullString
      'error
      blnError = True
      strErrMsg = LoadResString(4105)
      Exit Function
    End Select
    
    'get actual first arg
    intArg1 = GetNextArg(atVar, -1)
    'if error
    If blnError Then
      'adjust error message
      strErrMsg = LoadResString(4064)
      Exit Function
    End If
    
    intDir = 1
    'next character must be "="
    strArg2 = NextCommand()
    If strArg2 <> "=" Then
      'error
      blnError = True
      strErrMsg = LoadResString(4105)
      Exit Function
    End If
    
  'if this arg is string
  ElseIf ConvertArgument(strArg1, atStr) Then
    'string assignment
    '     s# = m#
    '     s# = "<string>"
    
    'get string variable number
    intArg1 = VariableValue(strArg1)
    
    'if outside expected bounds (strings should be limited to 0-23)
    If intArg1 > 23 Then
      Select Case agMainLogSettings.ErrorLevel
      Case leHigh
        'generate error
        blnError = True
        'for version 2.089, 2.272, and 3.002.149 only 12 strings
        Select Case agIntVersion
        Case 2.089, 2.272, 3.002149
          strErrMsg = Replace(Replace(LoadResString(4079), ARG1, "1"), ARG2, "11")
        Case Else
          strErrMsg = Replace(Replace(LoadResString(4079), ARG1, "1"), ARG2, "23")
        End Select
        Exit Function
        
      Case leMedium
        'generate warning
        'for version 2.089, 2.272, and 3.002.149 only 12 strings
        Select Case agIntVersion
        Case 2.089, 2.272, 3.002149
          AddWarning Replace(LoadResString(5007), ARG1, "11")
        Case Else
          AddWarning Replace(LoadResString(5007), ARG1, "11")
        End Select
      End Select
    End If
        
    'check for equal sign
    strArg2 = NextCommand()
    'if not equal sign
    If strArg2 <> "=" Then
      'error
      blnError = True
      strErrMsg = LoadResString(4034)
      Exit Function
    End If
    'get actual second variable
    'use argument extractor in case
    'second variable is a literal string)
    intArg2 = GetNextArg(atMsg, -1)
    'if error
    If blnError Then
      'just exit
      Exit Function
    End If
    
    'write set.string cmd
    bytCmd = &H72
  
  'if this is a variable
  ElseIf ConvertArgument(strArg1, atVar) Then
    'arg 1 must be 'v#' format
    intArg1 = VariableValue(strArg1)
    
    'if invalid variable number
    If intArg1 = -1 Then
      'invalid number
        blnError = True
        strErrMsg = LoadResString(4085)
      Exit Function
    End If
  
    'variable assignment or arithmetic operation
    'need next command to determine what kind of assignment/operation
    strArg2 = NextCommand()
    
    Select Case agCurrentSyntax
    Case lsVB
      'next argument should be equal sign
      If strArg2 <> "=" Then
        blnError = True
        strErrMsg = LoadResString(4034)
        Exit Function
      End If
    Case lsAGI
      Select Case strArg2
      Case "++"
        ' v#++;
        bytCmd = &H1
      Case "+="
        ' v# += #; or v# += v#;
        bytCmd = &H5
      Case "--"
        ' v#--
        bytCmd = &H2
      Case "-="
        ' v# -= #; or v# -= v#;
        bytCmd = &H7
      Case "*="
        ' v# *= #; or v# *= v#;
        bytCmd = &HA5
      Case "/="
        ' v# /= #; v# /= v#
        bytCmd = &HA7
      Case "="
        'right indirection
        '     v# = *v#;
        'assignment
        '     v# = v#;
        '     v# = #;
        'long arithmetic operation
        '     v# = v# + #; v# = v# + v#;
        '     v# = v# - #; v# = v# - v#;
        '     v# = v# * #; v# = v# * v#;
        '     v# = v# / #; v# = v# / v#;
      
      Case Else
        'don't know what the heck it is...
        blnError = True
        strErrMsg = LoadResString(4034)
        Exit Function
      End Select
    End Select
    
  'check for flag assignment
  ElseIf ConvertArgument(strArg1, atFlag) Then
    'flag assignment
    '     f# = True;
    '     f# = False;
    
    'get flag number
    intArg1 = VariableValue(strArg1)
    
    'check for equal sign
    strArg2 = NextCommand()
    'if not equal sign
    If strArg2 <> "=" Then
      'error
      blnError = True
      strErrMsg = LoadResString(4034)
      Exit Function
    End If
    
    'get flag value
    strArg2 = NextCommand()
    
    Select Case LCase$(strArg2)
    Case "true"
      'set this flag
      bytCmd = &HC
    
    Case "false"
      'reset this flag
      bytCmd = &HD
      
    Case Else
      'error
      blnError = True
      strErrMsg = LoadResString(4034)
      'always exit
      Exit Function
    End Select
  
  Else
    'not a special syntax
    CompileSpecial = False
    Exit Function
  End If
  
  'if not increment/decrement/set.string/reset/set
  Select Case bytCmd
  Case &H1, &H2, &HC, &HD, &H72
  Case Else
    'get next argument
    strArg2 = NextCommand()
    'if it is indirection
    If strArg2 = "*" Then
      'if not already using left indirection, AND cmd is not known
      If intDir = 0 And bytCmd = 0 Then
        'set right indirection
        intDir = 2
        
        'next char can't be a space, newline, or tab
        Select Case Mid$(strCurrentLine, lngPos + 1, 1)
        Case " ", vbTab, vbNullString
          'error
          blnError = True
          strErrMsg = LoadResString(4105)
          Exit Function
        End Select
        
        'get actual variable
        intArg2 = GetNextArg(atVar, -1)
        If blnError Then
          'reset error string
          strErrMsg = LoadResString(4105)
          Exit Function
        End If
      Else
        'error
        blnError = True
        strErrMsg = LoadResString(4105)
        Exit Function
      End If
    Else
      'arg2 is either number or variable- convert input to standard syntax
      blnArg2Var = True
      If Not ConvertArgument(strArg2, atNum, blnArg2Var) Then
        'set error
        blnError = True
        strErrMsg = LoadResString(4088)
        Exit Function
      End If
      
      intArg2 = VariableValue(strArg2)
      'if invalid
      If intArg2 = -1 Then
        strErrMsg = LoadResString(4088)
        Exit Function
      End If
      
      'if arg2 is a number
      If Not blnArg2Var Then
        'if cmd is not yet known,
        If bytCmd = 0 Then
          'must be assign
          ' v# = #
          ' *v# = #
          If intDir = 1 Then
            'lindirect.n
            bytCmd = &HB
          Else
            'assign.n
            bytCmd = &H3
          End If
        End If
      End If
    End If ' not indirection
  End Select 'if not inc/dec
  
  'if command is not known
  If bytCmd = 0 Then
    'if arg values are the same (already know arg2 is a variable)
    'and no indirection
    If (intArg1 = intArg2) And intDir = 0 Then
      'check for long arithmetic
      strArg2 = NextCommand()
      'if end of command is reached
      If ((agCurrentSyntax = lsAGI) And (strArg2 = ";")) Or ((agCurrentSyntax = lsVB) And (strArg2 = vbNullString)) Then
        'if agi syntax
        If agCurrentSyntax = lsAGI Then
          'move pointer back one space so eol
          'check in CompileAGI works correctly
          lngPos = lngPos - 1
        End If
        
        'this is a simple assign (with a variable being assigned to itself!!)
        Select Case agMainLogSettings.ErrorLevel
        Case leHigh
          blnError = True
          strErrMsg = LoadResString(4084)
          Exit Function
        Case leMedium
          AddWarning LoadResString(5036)
        Case leLow
          'allow it
        End Select
        bytCmd = &H3
      Else
        'this may be long arithmetic
        Select Case strArg2
        Case "+"
          bytCmd = &H5
        Case "-"
          bytCmd = &H7
        Case "*"
          bytCmd = &HA5
        Case "/"
          bytCmd = &HA7
        Case Else
          'error
          blnError = True
          strErrMsg = LoadResString(4087)
          Exit Function
        End Select
        
        'now get actual second argument
        blnArg2Var = True
        intArg2 = GetNextArg(atNum, -1, blnArg2Var)
        'if error
        If blnError Then
          'reset error msg
          strErrMsg = LoadResString(4088)
          Exit Function
        End If
      End If
    Else
      'variables are different
      'must be assignment
      ' v# = v#
      ' *v# = v#
      ' v# = *v#
      Select Case intDir
      Case 0  'assign.v
        bytCmd = &H4
      Case 1 'lindirect.v
        bytCmd = &H9
      Case 2  'rindirect
        bytCmd = &HA
        blnArg2Var = False
      End Select
      'always reset arg2var flag so
      'command won't be adjusted later
        blnArg2Var = False
    End If
  End If
  
  'check for special case of incrementing/decrementing
  'by adding/subtracting 1
  If bytCmd = &H5 And intArg2 = 1 And Not blnArg2Var Then
    'change command to increment
    bytCmd = &H1
  End If
  If bytCmd = &H7 And intArg2 = 1 And Not blnArg2Var Then
    'change to decrement
    bytCmd = &H2
  End If
  
  'if second argument is a variable
  If blnArg2Var Then
    bytCmd = bytCmd + 1
  End If
  
  'get next command on this line
  strArg2 = NextCommand(True)
  
  Select Case agCurrentSyntax
  Case lsVB
    'should be at end of line
    If LenB(strArg2) <> 0 Then
      blnError = True
      strErrMsg = LoadResString(4054)
      Exit Function
    End If
  Case lsAGI
    'check that next command is semicolon
    If strArg2 <> ";" Then
      blnError = True
      strErrMsg = LoadResString(4007)
      Exit Function
    Else
      'move pointer back one space so
      'eol check in CompileAGI works
      'correctly
      lngPos = lngPos - 1
    End If
  End Select
  'write command and arg1
  tmpLogRes.WriteByte bytCmd
  tmpLogRes.WriteByte CByte(intArg1)
  'write second argument for all cmds except &H1, &H2, &HC, &HD
  Select Case bytCmd
  Case &H1, &H2, &HC, &HD
  Case Else
    tmpLogRes.WriteByte CByte(intArg2)
  End Select
Exit Function

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 659, strErrSrc, Replace(LoadResString(659), ARG1, CStr(lngError) & ":" & strError)
End Function

Private Function LabelNum(LabelName As String) As Byte
  'this function will return the number of the label passed
  'as a string,
  'or zero, if a match is not found
  
  Dim i As Integer
  
  'step through all labels,
  For i = 1 To bytLabelCount
    If llLabel(i).Name = LabelName Then
      LabelNum = i
      Exit Function
    End If
  Next i
  
  'if not found, zero is returned
End Function
